import os
import sys
import csv
import anndata as ad
import pandas as pd
import numpy as np
from scipy.sparse import csr_matrix
from scipy.io import mmread
import collections
sys.path.insert(1, os.path.join(sys.path[0], '..'))
import long_read.isoform_junctions as junc
import long_read.isoform_matrix as iso
import long_read.isoform_ratios as isor
import long_read.isoform_translation as isot
import long_read.gff_process as gff_process
from Bio import SeqIO
from oncosplice.metadataAnalysis import *

def exportRatios(ob,out):
    pseudo_pdf, tpm, isoform_to_gene_ratio = iso.pseudo_cluster_counts(ob,cell_threshold=5,count_threshold=0,compute_tpm=True)
    pseudo_pdf.to_csv(out+'_pseudo_cluster_counts.txt', sep='\t')
    tpm.to_csv(out+'_pseudo_cluster-tpm.txt', sep='\t')
    isoform_to_gene_ratio.to_csv(out+'_pseudo_cluster-ratio.txt', sep='\t')

# Reference transcriptome
ensembl_exon_dir = '/Users/saljh8/Documents/GitHub/altanalyze/AltDatabase/EnsMart91/ensembl/Hs/Hs_Ensembl_exon.txt'

############# GTF Files ###############
# Reference annotations (note - the program saves results automatically witht the GTF name)
gencode_gff = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/gencode.v45.annotation.gff3'
gff_RefSeq = '/Users/saljh8/Dropbox/Revio/ENCODE/Aggregate/hg38.ncbiRefSeq.gtf'
gff_GTEx = '/Users/saljh8/Dropbox/Revio/GTEx/flair_filter_transcripts.gtf'
# Young
gff_wm34 = '/Users/saljh8/Dropbox/Revio/Young-Healthy-Revio/PacBio/WM34/WM34.gff'
gff_bm27 = '/Users/saljh8/Dropbox/Revio/Young-Healthy-Revio/PacBio/BM27/BM27.gff'
gff_bf21 = '/Users/saljh8/Dropbox/Revio/Young-Healthy-Revio/PacBio/BF21/BF21.gff'
gff_167h = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/HSC_3k/ND167_HSC-3k.gtf'
gff_167m = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/MPP_3k/ND167_MPP_3k.gtf'
gff_167l = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/LMPP_3k/ND167_LMPP_3k.gtf'
gff_251_1 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/PacBio/first_run/ND251-1.gff'
gff_251_2 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/PacBio/size_selected/ND251-2.gff'
# Aging
gff_am72 = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/AM72/AM72.gff'
gff_wf83 = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/WF83/WF83.gff'
gff_bf71 = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/BF71/BF71.gff'
# MDS
gff_5801M_pre ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/MDS-pre/5801-pre.gff'
gff_5801M_post ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/MDS-post/5801-post.gff'
gff_5801A_pre1 = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML/run1/5801-AML-1.gff'
gff_5801A_pre2 ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML/run2/5801-AML-2.gff'
gff_5801A_therapy ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML-therapy/5801IR.gff'
gff_5801A_post ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML-post/5801-9_341.gff'
gff_am75M_pre ='/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/pre/Y1258.gff'
gff_am75M_therapy ='/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/on/Y1607.gff'
gff_am75M_post ='/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/post/Y5116.gff'
gff_wf78M_pre ='/Users/saljh8/Dropbox/Revio/MDS-Revio/WF78-MDS/pre/Y1493.gff'
gff_wf78M_post ='/Users/saljh8/Dropbox/Revio/MDS-Revio/WF78-MDS/post/Y1737.gff'

############# Matrix Directories ###############
# Young
matrix_wm34 = '/Users/saljh8/Dropbox/Revio/Young-Healthy-Revio/PacBio/WM34/isoforms_seurat'
matrix_bm27 = '/Users/saljh8/Dropbox/Revio/Young-Healthy-Revio/PacBio/BM27/isoforms_seurat'
matrix_bf21 = '/Users/saljh8/Dropbox/Revio/Young-Healthy-Revio/PacBio/BF21/isoforms_seurat'
matrix_167h = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/HSC_3k/filtered_matrices'
matrix_167m = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/MPP_3k/filtered_matrices'
matrix_167l = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/LMPP_3k/filtered_matrices'
matrix_251_1 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/PacBio/first_run/scisoseq.seurat_info/isoforms_seurat'
matrix_251_2 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/PacBio/size_selected/scisoseq.seurat_info/isoforms_seurat'
# Aging
matrix_am72 = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/AM72/scisoseq.seurat_info/isoforms_seurat'
matrix_wf83 = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/WF83/scisoseq.seurat_info/isoforms_seurat'
matrix_bf71 = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/BF71/isoforms_seurat'
# MDS
matrix_5801M_pre ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/MDS-pre/isoforms_seurat'
matrix_5801M_post ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/MDS-post/isoforms_seurat'
matrix_5801A_pre1 = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML/run1/isoforms_seurat'
matrix_5801A_pre2 ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML/run2/isoforms_seurat'
matrix_5801A_therapy ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML-therapy/isoforms_seurat'
matrix_5801A_post ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML-post/isoforms_seurat'
matrix_am75M_pre ='/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/pre/isoforms_seurat'
matrix_am75M_therapy ='/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/on/isoforms_seurat'
matrix_am75M_post ='/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/post/isoforms_seurat'
matrix_wf78M_pre ='/Users/saljh8/Dropbox/Revio/MDS-Revio/WF78-MDS/pre/isoforms_seurat'
matrix_wf78M_post ='/Users/saljh8/Dropbox/Revio/MDS-Revio/WF78-MDS/post/isoforms_seurat'

bc_dir1 = '/Users/saljh8/Dropbox/Revio/Young-Healthy-Revio/Illumina/SoupX-hg38/cellHarmony-labels-young-old.txt'
bc_dir2 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/QueryGroups.cellHarmony.txt'
bc_dir3 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/Illumina/SoupX-hg38/cellHarmony/QueryGroups.cellHarmony.txt'
bc_dir4 = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/Illumina/SoupX-hg38/cellHarmony/QueryGroups.cellHarmony.txt'
bc_dir5 = '/Users/saljh8/Dropbox/Revio/MDS-Revio/Aggregate/QueryGroups.cellHarmony.txt'
barcode_cluster_dirs = [bc_dir1,bc_dir2,bc_dir3,bc_dir4,bc_dir5]
barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dirs)

healthy_junction_counts = False
if healthy_junction_counts:

    ######## Process data from Young Bone Marrow samples ########
    ### Requires 256GB of RAM
    adata = junc.exportJunctionMatrix(matrix_wm34, ensembl_exon_dir, gff_wm34, barcode_clusters=barcode_sample_dict['WM34_CD34'],rev=True)
    adata = junc.exportJunctionMatrix(matrix_bf21, ensembl_exon_dir, gff_bf21, barcode_clusters=barcode_sample_dict['BF21-CD34'],rev=True)
    adata = junc.exportJunctionMatrix(matrix_bm27, ensembl_exon_dir, gff_bm27, barcode_clusters=barcode_sample_dict['BM27_CD34'],rev=True)
    
    hsc_nd167_adata = junc.exportJunctionMatrix(matrix_167h, ensembl_exon_dir, gff_167h, barcode_clusters=barcode_sample_dict['WM35_ND20-167__HSC_3k'])
    mpp_nd167_adata = junc.exportJunctionMatrix(matrix_167m, ensembl_exon_dir, gff_167m, barcode_clusters=barcode_sample_dict['WM35_ND20-167__MPP_3k'])
    lmpp_nd167_adata = junc.exportJunctionMatrix(matrix_167l, ensembl_exon_dir, gff_167l, barcode_clusters=barcode_sample_dict['WM35_ND20-167__LMPP_3k'])
    
    # Ensure each has adata has capture specific cell barcodes and combine
    hsc_nd167_adata = iso.append_sample_name(hsc_nd167_adata, 'WM35_ND20-167__HSC_3k')
    mpp_nd167_adata = iso.append_sample_name(mpp_nd167_adata, 'WM35_ND20-167__MPP_3k')
    lmpp_nd167_adata = iso.append_sample_name(lmpp_nd167_adata, 'WM35_ND20-167__LMPP_3k')
    nd167_adata = ad.concat([hsc_nd167_adata, mpp_nd167_adata, lmpp_nd167_adata], axis=0, join='outer')
    nd167_adata.write_h5ad('ND167.h5ad')
    
    nd251_1_adata = junc.exportJunctionMatrix(matrix_251_1, ensembl_exon_dir, gff_251_1, barcode_clusters=barcode_sample_dict['WF40_ND21-251_HSC_3k'],rev=True)
    nd251_2_adata = junc.exportJunctionMatrix(matrix_251_2, ensembl_exon_dir, gff_251_2, barcode_clusters=barcode_sample_dict['WF40_ND21-251_HSC_3k'],rev=True)
    nd251_1_adata = iso.append_sample_name(nd251_1_adata, 'WF40_ND21-251_HSC_3k-1')
    nd251_2_adata = iso.append_sample_name(nd251_2_adata, 'WF40_ND21-251_HSC_3k-2')
    nd251_adata = ad.concat([nd251_1_adata, nd251_2_adata], axis=0, join='outer')
    nd251_adata.write_h5ad('ND251-apharesis.h5ad')
    
    ######## Process data from elderly Bone Marrow samples ########

    # Produce an junction counts h5ad for the AM72-N-0206_CD34
    junc.exportJunctionMatrix(matrix_am72, ensembl_exon_dir, gff_am72, barcode_clusters=barcode_sample_dict['AM72-N-0206_CD34'],rev=True)
    junc.exportJunctionMatrix(matrix_wf83, ensembl_exon_dir, gff_wf83, barcode_clusters=barcode_sample_dict['WF83-N-0215_CD34'],rev=True)
    junc.exportJunctionMatrix(matrix_bf71, ensembl_exon_dir, gff_bf71, barcode_clusters=barcode_sample_dict['BF71-N-0421_CD34'],rev=True)
    
    ######## Process data from MDS Bone Marrow samples ########

    adata = junc.exportJunctionMatrix(matrix_5801M_pre,ensembl_exon_dir,gff_5801M_pre,barcode_clusters=barcode_sample_dict['WM71-0121-5801_CD34'],rev=True)
    adata = junc.exportJunctionMatrix(matrix_5801M_post,ensembl_exon_dir,gff_5801M_post,barcode_clusters=barcode_sample_dict['MDS5801-postHMA_CD34'],rev=True)

    p5801_aml_adata_1 = junc.exportJunctionMatrix(matrix_5801A_pre1,ensembl_exon_dir,gff_5801A_pre1,barcode_clusters=barcode_sample_dict['MDS5801-preVEN_342'],rev=True)
    p5801_aml_adata_2 = junc.exportJunctionMatrix(matrix_5801A_pre2,ensembl_exon_dir,gff_5801A_pre2,barcode_clusters=barcode_sample_dict['MDS5801-preVEN_342'],rev=True,)
    
    nd251_1_adata = ad.read_h5ad('ND251-1.h5ad')
    nd251_2_adata = ad.read_h5ad('ND251-2.h5ad')

    # Ensure each has adata has capture specific cell barcodes and combine
    nd251_1_adata = iso.append_sample_name(nd251_1_adata, 'WF40_ND21-251_HSC_3k-1')
    nd251_2_adata = iso.append_sample_name(nd251_2_adata, 'WF40_ND21-251_HSC_3k-2')
    nd251_adata = ad.concat([nd251_1_adata, nd251_2_adata], axis=0, join='outer')
    nd251_adata.write_h5ad('5801-AML.h5ad')

    p5801_aml_therapy_adata = junc.exportJunctionMatrix(matrix_5801A_therapy,ensembl_exon_dir,gff_5801A_therapy,barcode_sample_dict['MDS5801-1229IR_CD34'],rev=True)
    p5801_aml_post_adata = junc.exportJunctionMatrix(matrix_5801A_post,ensembl_exon_dir,gff_5801A_post,barcode_clusters=barcode_sample_dict['MDS5801R-341'],rev=True)

    am75_post_adata = junc.exportJunctionMatrix(matrix_am75M_post,ensembl_exon_dir,gff_source=gff_am75M_post,barcode_clusters=barcode_sample_dict['Y5116'],rev=True)
    am75_pre_adata = junc.exportJunctionMatrix(matrix_am75M_pre,ensembl_exon_dir,gff_source=gff_am75M_pre,barcode_clusters=barcode_sample_dict['Y1258'],rev=True)
    am75_on_adata = junc.exportJunctionMatrix(matrix_am75M_therapy,ensembl_exon_dir,gff_source=gff_am75M_therapy,barcode_clusters=barcode_sample_dict['Y1607'],rev=True)
    wf78_post_adata = junc.exportJunctionMatrix(matrix_wf78M_post,ensembl_exon_dir,gff_source=gff_wf78M_post,barcode_clusters=barcode_sample_dict['Y1737'],rev=True)
    wf78_pre_adata = junc.exportJunctionMatrix(matrix_wf78M_pre,ensembl_exon_dir,gff_source=gff_wf78M_pre,barcode_clusters=barcode_sample_dict['Y1493'],rev=True)

aggregate_junction_data = True
if aggregate_junction_data:
    # Define the list of sample files
    sample_files = [
        'WM34.h5ad','BM27.h5ad','BF21.h5ad','ND251-apharesis.h5ad','ND167.h5ad',
        'BF71.h5ad','WF83.h5ad','AM72.h5ad',
        '5801-pre','5801-post', '5801-AML', '5801IR','5801-9_341',
        'Y1258','Y1607','Y5116','Y1493','Y1737'
    ]

    pseudo_counts,pseudo_tpm,pseudo_ratios = iso.concatenate_h5ad_and_compute_pseudobulks(sample_files,collection_name='Old-Young')

    # Filter and re-order exported pseudobulks
    input_file = pseudo_counts
    output_file = 'filtered_combined_junction_counts.txt'
    groups_file = 'groups_file2.txt'
    cell_type_order_file = 'cell_type_order.txt'  # Optional
    iso.export_and_filter_pseudobulks(input_file, output_file, groups_file, cell_type_order_file)


    #### Run AltAnalyze2 with min_exp=4;med_exp=9, as
    # python ExpressionBuilder.py --i /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/AltAnalyze2/ --a unbiased --species Hs --platform RNASeq --additional /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/AltAnalyze2/ExpressionInput/counts.ND-junctions.txt
    # python import_scripts/AugmentEventAnnotations.py --i /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/AltAnalyze2/AltResults/AlternativeOutput/Hs_RNASeq_top_alt_junctions-PSI.txt --s Hs --platform RNASeq
    # python stats_scripts/metaDataAnalysis.py --i /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/AltAnalyze2/AltResults/AlternativeOutput/Hs_RNASeq_top_alt_junctions-PSI_EventAnnotation.txt --s Hs --p PSI --dPSI 0.1 --percentExp 40 --m /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/AltAnalyze2/ExpressionInput/groups.ND-junctions.txt

################################### ISOFORM ANNOTATION ###################################

build_healthy_iso_ref = False
if build_healthy_iso_ref:
    #### Create a combined GFF and find the longest consensus isoforms
    gff_source =  [gencode_gff,gff_RefSeq,gff_GTEx]
    gff_source += [gff_wm34,gff_bm27,gff_bf21,gff_167h,gff_167m,gff_167l,gff_251_1,gff_251_2]
    gff_source += [gff_am72,gff_wf83,gff_bf71]
    gff_source += [gff_5801M_pre,gff_5801M_post,gff_5801A_pre1,gff_5801A_pre2,gff_5801A_therapy,gff_5801A_post]
    gff_source += [gff_am75M_pre,gff_am75M_therapy,gff_am75M_post,gff_am75M_therapy,gff_am75M_post,gff_wf78M_pre,gff_wf78M_post]
    gff_output_dir = gff_process.consolidateLongReadGFFs(gff_source, ensembl_exon_dir,mode='Ensembl')

    #### Translate isoform into protein sequence and check for NMD
    ref_gff_file = "/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/gencode.v45.annotation.gff3"
    query_gff_file = gff_output_dir+"/combined.gff"
    genome_fasta = "/Users/saljh8/Dropbox/Revio/Other/Variants/SNV/genome.fa"
    transcript_associations_file = gff_output_dir+"/transcript_associations.txt"
    cds_records, transcript_records, protein_records = isot.gff_translate(query_gff_file,genome_fasta,ref_gff_file,transcript_associations_file)

    with open("protein_sequences.fasta", "w") as protein_file:
        SeqIO.write(protein_records, protein_file, "fasta")

    with open("cds_sequences.fasta", "w") as cds_file:
        SeqIO.write(cds_records, cds_file, "fasta")
    sys.exit()

integrate_healthy_malignant_isoforms = False
if integrate_healthy_malignant_isoforms:
    print("Integrating malignant and healthy for pseudobulk cell populations...")

    ######## Process data from Young Bone Marrow samples ########
    isoform_association_path = "/Users/saljh8/Dropbox/Revio/BoneMarrow/gff-output/isoform_links.txt"

    wm34_adata = isor.exportConsensusIsoformMatrix(matrix_wm34,isoform_association_path,gff_source=gff_wm34,barcode_clusters=barcode_sample_dict['WM34_CD34'],rev=True)
    bf21_adata = isor.exportConsensusIsoformMatrix(matrix_bf21,isoform_association_path,gff_source=gff_bf21,barcode_clusters=barcode_sample_dict['BF21-CD34'],rev=True)
    bm27_adata = isor.exportConsensusIsoformMatrix(matrix_bm27,isoform_association_path,gff_source=gff_bm27,barcode_clusters=barcode_sample_dict['BM27_CD34'],rev=True)
    hsc_nd167_adata = isor.exportConsensusIsoformMatrix(matrix_167h,isoform_association_path,gff_source=gff_167h,barcode_clusters=barcode_sample_dict['WM35_ND20-167__HSC_3k'],sample_id='ND20-167')
    mpp_nd167_adata = isor.exportConsensusIsoformMatrix(matrix_167m,isoform_association_path,gff_source=gff_167m,barcode_clusters=barcode_sample_dict['WM35_ND20-167__MPP_3k'],sample_id='ND20-167')
    lmpp_nd167_adata = isor.exportConsensusIsoformMatrix(matrix_167l,isoform_association_path,gff_source=gff_167l,barcode_clusters=barcode_sample_dict['WM35_ND20-167__LMPP_3k'],sample_id='ND20-167')
    hsc_nd167_adata = iso.append_sample_name(hsc_nd167_adata, 'WM35_ND20-167__HSC_3k')
    mpp_nd167_adata = iso.append_sample_name(mpp_nd167_adata, 'WM35_ND20-167__MPP_3k')
    lmpp_nd167_adata = iso.append_sample_name(lmpp_nd167_adata, 'WM35_ND20-167__LMPP_3k')
    nd167_adata = ad.concat([hsc_nd167_adata, mpp_nd167_adata, lmpp_nd167_adata], axis=0, join='outer')
    nd167_adata.write_h5ad('ND167-isoform.h5ad')
    nd251_1_adata = isor.exportConsensusIsoformMatrix(matrix_251_1,isoform_association_path,gff_source=gff_251_1,rev=True,barcode_clusters=barcode_sample_dict['WF40_ND21-251_HSC_3k'],sample_id='ND251')
    nd251_2_adata = isor.exportConsensusIsoformMatrix(matrix_251_1,isoform_association_path,gff_source=gff_251_2,rev=True,barcode_clusters=barcode_sample_dict['WF40_ND21-251_HSC_3k'],sample_id='ND251')
    nd251_1_adata = iso.append_sample_name(nd251_1_adata, 'WF40_ND21-251_HSC_3k-1')
    nd251_2_adata = iso.append_sample_name(nd251_2_adata, 'WF40_ND21-251_HSC_3k-2')
    nd251_adata = ad.concat([nd251_1_adata, nd251_2_adata], axis=0, join='outer')
    nd251_adata.write_h5ad('ND251-apharesis-isoform.h5ad')

    ######## Process data from Elderly Bone Marrow samples ########

    bf71_adata = isor.exportConsensusIsoformMatrix(matrix_bf71,isoform_association_path,gff_source=gff_bf71,barcode_clusters=barcode_sample_dict['BF71-N-0421_CD34'],rev=True)



    ######## Process data from elderly Bone Marrow samples ########
    barcode_cluster_dir = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/Illumina/SoupX-hg38/cellHarmony/QueryGroups.cellHarmony.txt'
    barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dir)
    
    # Produce an junction counts h5ad for the AM72-N-0206_CD34
    matrix_path = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/AM72/scisoseq.seurat_info/isoforms_seurat'
    gff_source = 'AM72'
    am72_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,barcode_clusters=barcode_sample_dict['AM72-N-0206_CD34'],rev=True)

    # Produce an junction counts h5ad for the WF83-N-0215_CD34 capture
    matrix_path = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/WF83/scisoseq.seurat_info/isoforms_seurat'
    gff_source = 'WF83'
    wf83_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,barcode_clusters=barcode_sample_dict['WF83-N-0215_CD34'],rev=True)

    barcode_cluster_dir = '/Users/saljh8/Dropbox/Revio/MDS-Revio/Aggregate/QueryGroups.cellHarmony.txt'
    barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dir)
    
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML/run1/isoforms_seurat'
    gff_source = '5801-AML-1'
    p5801_aml_adata_1 = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,barcode_clusters=barcode_sample_dict['MDS5801-preVEN_342'],rev=True,sample_id='5801')
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML/run2/isoforms_seurat'
    gff_source = '5801-AML-2'
    p5801_aml_adata_2 = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,barcode_clusters=barcode_sample_dict['MDS5801-preVEN_342'],rev=True,sample_id='5801')
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML-therapy/isoforms_seurat'
    gff_source = '5801IR'
    p5801_aml_therapy_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,barcode_clusters=barcode_sample_dict['MDS5801-1229IR_CD34'],rev=True,sample_id='5801')
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML-post/isoforms_seurat'
    gff_source = '5801-9_341'
    p5801_aml_post_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['MDS5801R-341'],sample_id='5801')
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/MDS-post/isoforms_seurat'
    gff_source = '5801-post'
    p5801_mds_post_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['MDS5801-postHMA_CD34'],sample_id='5801')
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/MDS-pre/isoforms_seurat'
    gff_source = '5801-pre'
    p5801_mds_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['WM71-0121-5801_CD34'],sample_id='5801')
    p5801_aml_adata_1 = iso.append_sample_name(p5801_aml_adata_1, '5801-AML-1')
    p5801_aml_adata_2 = iso.append_sample_name(p5801_aml_adata_2, '5801-AML-2')
    p5801_aml_therapy_adata = iso.append_sample_name(p5801_aml_therapy_adata, '5801IR')
    p5801_aml_post_adata = iso.append_sample_name(p5801_aml_post_adata, '5801-9_341')
    p5801_mds_post_adata = iso.append_sample_name(p5801_mds_post_adata, '5801-post')
    p5801_mds_adata = iso.append_sample_name(p5801_mds_adata, '5801-pre')
    p5801_adata = ad.concat([p5801_aml_adata_1, p5801_aml_adata_2,p5801_aml_therapy_adata,p5801_aml_post_adata,p5801_mds_post_adata,p5801_mds_adata], axis=0, join='outer')
    p5801_adata.write_h5ad('5801.h5ad')

    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/post/isoforms_seurat'
    gff_source = 'Y5116' #AM75-MDS-post
    am75_post_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['Y5116'],sample_id='AM75')
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/pre/isoforms_seurat'
    gff_source = 'Y1258' # AM75-MDSpre
    am75_pre_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['Y1258'],sample_id='AM75')
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/on/isoforms_seurat'
    gff_source = 'Y1607' #AM75-MDSpre
    am75_on_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['Y1607'],sample_id='AM75')
    am75_post_adata = iso.append_sample_name(am75_post_adata, 'Y5116')
    am75_pre_adata = iso.append_sample_name(am75_pre_adata, 'Y1258')
    am75_on_adata = iso.append_sample_name(am75_on_adata, 'Y1607')
    am75_adata = ad.concat([am75_post_adata, am75_pre_adata, am75_on_adata], axis=0, join='outer')
    am75_adata.write_h5ad('am75.h5ad')

    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/WF78-MDS/post/isoforms_seurat'
    gff_source = 'Y1737' #WF78-MDS-post
    wf78_post_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['Y1737'],sample_id='WF78')
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/WF78-MDS/pre/isoforms_seurat'
    gff_source = 'Y1493' #WF78-MDS-pre
    wf78_pre_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['Y1493'],sample_id='WF78')
    wf78_post_adata = iso.append_sample_name(wf78_post_adata, 'Y1737')
    wf78_pre_adata = iso.append_sample_name(wf78_pre_adata, 'Y1493')
    wf78_adata = ad.concat([wf78_post_adata, wf78_pre_adata], axis=0, join='outer')
    wf78_adata.write_h5ad('wf78.h5ad')

    barcode_cluster_dir = '/Users/saljh8/Dropbox/Revio/AML-Revio/QueryGroups.cellHarmony.txt'
    barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dir)
    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML14/run1/isoforms_seurat'
    gff_source = 'AML14'
    aml14_1_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['AML-14_CITE_GEX'],sample_id='AML14')
    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML14/run2/isoforms_seurat'
    gff_source = 'AML14-2'
    aml14_2_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['AML-14_CITE_GEX'],sample_id='AML14')
    aml14_1_adata = iso.append_sample_name(aml14_1_adata, 'AML14')
    aml14_2_adata = iso.append_sample_name(aml14_2_adata, 'AML14-2')
    aml14_adata = ad.concat([aml14_1_adata, aml14_2_adata], axis=0, join='outer')
    aml14_adata.write_h5ad('aml14.h5ad')

    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/12-run1/isoforms_seurat'
    gff_source = 'AML12'
    aml12_1_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['AML-12_1_CITE_GEX'],sample_id='AML12')
    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/12-run2/isoforms_seurat'
    gff_source = 'AML12-2'
    aml12_2_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['AML-12_1_CITE_GEX'],sample_id='AML12')
   
    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/13-run1/isoforms_seurat'
    gff_source = 'AML13'
    aml12_3_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['AML-13_CITE_GEX'],sample_id='AML12')
    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/13-run2/isoforms_seurat'
    gff_source = 'AML13-2'
    aml12_4_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['AML-13_CITE_GEX'],sample_id='AML12')
    aml12_1_adata = iso.append_sample_name(aml12_1_adata, 'AML12')
    aml12_2_adata = iso.append_sample_name(aml12_2_adata, 'AML12-2')
    aml12_3_adata = iso.append_sample_name(aml12_3_adata, 'AML13')
    aml12_4_adata = iso.append_sample_name(aml12_4_adata, 'AML13-2')
    aml12_adata = ad.concat([aml12_1_adata, aml12_2_adata,aml12_3_adata,aml12_4_adata], axis=0, join='outer')
    aml12_adata.write_h5ad('aml12.h5ad')

    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML7/run1/isoforms_seurat'
    gff_source = 'AML7'
    aml7_1_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['AML-7_1_CITE_GEX'],sample_id='AML7')
    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML7/run2/isoforms_seurat'
    gff_source = 'AML7-2'
    aml7_2_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,rev=True,barcode_clusters=barcode_sample_dict['AML-7_1_CITE_GEX'],sample_id='AML7')
    aml7_1_adata = iso.append_sample_name(aml7_1_adata, 'AML7')
    aml7_2_adata = iso.append_sample_name(aml7_2_adata, 'AML7-2')
    aml7_adata = ad.concat([aml7_1_adata, aml7_2_adata], axis=0, join='outer')
    aml7_adata.write_h5ad('aml7.h5ad')
    
    am75_pre_adata = ad.read_h5ad('Y1258-isoform.h5ad')
    am75_post_adata = ad.read_h5ad('Y5116-isoform.h5ad')
    am75_post_adata = iso.append_sample_name(am75_post_adata, 'Y5116')
    am75_pre_adata = iso.append_sample_name(am75_pre_adata, 'Y1258')
    am75_adata = ad.concat([am75_pre_adata, am75_post_adata], axis=0, join='outer')
    am75_adata.write_h5ad('am75-pre-post.h5ad')
    exportRatios(am75_adata,'am75-pre-post')
    
    
    p5801_aml_post_adata = ad.read_h5ad('5801-9_341-isoform.h5ad')
    p5801_mds_post_adata = ad.read_h5ad('5801-post-isoform.h5ad')
    p5801_mds_adata = ad.read_h5ad('5801-pre-isoform.h5ad')
    p5801_aml_therapy_adata = ad.read_h5ad('5801IR-isoform.h5ad')
    p5801_aml_adata_2 = ad.read_h5ad('5801-AML-2-isoform.h5ad')
    p5801_aml_adata_1 = ad.read_h5ad('5801-AML-1-isoform.h5ad')

    p5801_aml_adata_1 = iso.append_sample_name(p5801_aml_adata_1, '5801-AML-1')
    p5801_aml_adata_2 = iso.append_sample_name(p5801_aml_adata_2, '5801-AML-2')
    p5801_aml_therapy_adata = iso.append_sample_name(p5801_aml_therapy_adata, '5801IR')
    p5801_aml_post_adata = iso.append_sample_name(p5801_aml_post_adata, '5801-9_341')
    p5801_mds_post_adata = iso.append_sample_name(p5801_mds_post_adata, '5801-post')
    p5801_mds_adata = iso.append_sample_name(p5801_mds_adata, '5801-pre')
    p5801_adata = ad.concat([p5801_aml_therapy_adata,p5801_aml_post_adata], axis=0, join='outer')
    p5801_adata.write_h5ad('5801-AML-therapy.h5ad')
    exportRatios(p5801_adata,'5801-AML-therapy')

    #nd167_adata = ad.read_h5ad('ND167-isoform.h5ad');print('ND167-isoform.h5ad')
    #nd251_adata = ad.read_h5ad('ND251-apharesis-isoform.h5ad')

    """

    am72_adata = ad.read_h5ad('AM72-isoform.h5ad'); print('AM72-isoform.h5ad')
    wf83_adata = ad.read_h5ad('WF83-isoform.h5ad'); print('WF83-isoform.h5ad')
    wm34_adata = ad.read_h5ad('WM34-isoform.h5ad')
    bf71_adata = ad.read_h5ad('BF71-isoform.h5ad')
    bf21_adata = ad.read_h5ad('BF21-isoform.h5ad')
    bm27_adata = ad.read_h5ad('BM27-isoform.h5ad')

    # Make sure the indices are strings
    wm34_adata.obs.index = wm34_adata.obs.index.astype(str)
    bf21_adata.obs.index = bf21_adata.obs.index.astype(str)
    bf71_adata.obs.index = bf71_adata.obs.index.astype(str)
    bm27_adata.obs.index = bm27_adata.obs.index.astype(str)
    wf83_adata.obs.index = wf83_adata.obs.index.astype(str)
    am72_adata.obs.index = am72_adata.obs.index.astype(str)

    # Make observation names unique to avoid duplication issues
    wm34_adata.obs_names_make_unique()
    bf21_adata.obs_names_make_unique()
    bf71_adata.obs_names_make_unique()
    bm27_adata.obs_names_make_unique()
    wf83_adata.obs_names_make_unique()
    am72_adata.obs_names_make_unique()
    
    #### WARNING - MEMORY INTENSIVE - NOT RECOMMENDED - USE BELOW
    combined_adata = ad.concat([wm34_adata,bf21_adata,bm27_adata,bf71_adata,am72_adata, wf83_adata], axis=0, join='outer')

    pseudo_pdf, tpm, isoform_to_gene_ratio = iso.pseudo_cluster_counts(combined_adata,cell_threshold=5,count_threshold=0,compute_tpm=True)
    pseudo_pdf.to_csv('ND-MDS-AML-combined_pseudo_cluster_counts.txt', sep='\t')
    tpm.to_csv('ND-MDS-AM-combined_pseudo_cluster-tpm.txt', sep='\t')
    isoform_to_gene_ratio.to_csv('ND-MDS-AM-combined_pseudo_cluster-ratio.txt', sep='\t')
    """

    # Define the list of sample files
    sample_files = [
        'WM34-isoform.h5ad',
        'BF21-isoform.h5ad',
        'BF71-isoform.h5ad',
        'BM27-isoform.h5ad',
        'WF83-isoform.h5ad',
        'AM72-isoform.h5ad',
        'ND251-apharesis-isoform.h5ad',
        'ND167-isoform.h5ad'
    ]

    # Memory efficient combine and pseudobulk of many h5ad files
    pseudo_counts,pseudo_tpm,pseudo_ratios = concatenate_h5ad_and_compute_pseudobulks(sample_files,collection_name='ND-MDS-AML')

    # Filter and re-order exported pseudobulks
    input_file = pseudo_tpm
    input_file = pseudo_ratios
    input_file = 'filtered_ratio-filtered.txt'
    output_file = 'filtered_ratio2.txt'
    groups_file = 'groups_file2.txt'
    cell_type_order_file = 'cell_type_order.txt'  # Optional

    iso.export_and_filter_pseudobulks(input_file, output_file, groups_file, cell_type_order_file)




run_healthy_marrow_iso = False
if run_healthy_marrow_iso:

    ################################### ISOFORM LEVEL ANALYSIS ###################################
    #### Further collapse the isoform matrix for each sample using multi-sample consensus IDs

    # Indicates the "best" isoform across smaples
    isoform_association_path = "/Users/saljh8/Dropbox/Revio/BoneMarrow/gff-output/isoform_links.txt"

    # Import the cellHarmony results and save cell type pd frame for each cell barcode for each separate capture
    barcode_cluster_dir = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/QueryGroups.cellHarmony.txt'
    barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dir)

    # Produce an isoform counts h5ad for the WM35_ND20-167__HSC_3k capture
    matrix_path = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/HSC_3k/filtered_matrices'
    gff_source = 'ND167_HSC-3k'
    hsc_nd167_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,barcode_clusters=barcode_sample_dict['WM35_ND20-167__HSC_3k'])

    # Produce an junction counts h5ad for the WM35_ND20-167__MPP_3k capture
    matrix_path = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/MPP_3k/filtered_matrices'
    gff_source = 'ND167_MPP_3k'
    mpp_nd167_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,barcode_clusters=barcode_sample_dict['WM35_ND20-167__MPP_3k'])

    # Produce an junction counts h5ad for the WM35_ND20-167__LMPP_3k capture
    matrix_path = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/LMPP_3k/filtered_matrices'
    gff_source = 'ND167_LMPP_3k'
    lmpp_nd167_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,barcode_clusters=barcode_sample_dict['WM35_ND20-167__LMPP_3k'])

    # Combine and export pseudobulk results
    #hsc_nd167_adata = ad.read_h5ad('ND167_HSC-3k-isoform.h5ad')
    #mpp_nd167_adata = ad.read_h5ad('ND167_MPP-3k-isoform.h5ad')
    #lmpp_nd167_adata = ad.read_h5ad('ND167_LMPP-3k-isoform.h5ad')

    # Ensure each has adata has capture specific cell barcodes and combine
    hsc_nd167_adata = iso.append_sample_name(hsc_nd167_adata, 'WM35_ND20-167__HSC_3k')
    mpp_nd167_adata = iso.append_sample_name(mpp_nd167_adata, 'WM35_ND20-167__MPP_3k')
    lmpp_nd167_adata = iso.append_sample_name(lmpp_nd167_adata, 'WM35_ND20-167__LMPP_3k')
    nd167_adata = ad.concat([hsc_nd167_adata, mpp_nd167_adata, lmpp_nd167_adata], axis=0, join='outer')
    nd167_adata.write_h5ad('ND167-isoform.h5ad')
    pseudo_pdf, tpm, isoform_to_gene_ratio = iso.pseudo_cluster_counts(nd167_adata,cell_threshold=5,count_threshold=0,compute_tpm=True)
    pseudo_pdf.to_csv('nd167-combined_pseudo_cluster_counts.txt', sep='\t')
    tpm.to_csv('nd167-combined_pseudo_cluster-tpm.txt', sep='\t')
    isoform_to_gene_ratio.to_csv('nd167-combined_pseudo_cluster-ratio.txt', sep='\t')

    ######## Process data from sample ND251 ########
    # Import the cellHarmony results and save cell type pd frame for each cell barcode for each separate capture
    barcode_cluster_dir = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/Illumina/SoupX-hg38/cellHarmony/QueryGroups.cellHarmony.txt'
    barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dir)

    # Produce an junction counts h5ad for the WF40_ND21-251_HSC_3k capture (first run)
    matrix_path = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/PacBio/first_run/scisoseq.seurat_info/isoforms_seurat'
    gff_source = 'ND251-1'
    nd251_1_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,barcode_clusters=barcode_sample_dict['WF40_ND21-251_HSC_3k'],rev=True)

    # Produce an junction counts h5ad for the WF40_ND21-251_HSC_3k capture (second run)
    matrix_path = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/PacBio/size_selected/scisoseq.seurat_info/isoforms_seurat'
    gff_source = 'ND251-2'
    nd251_2_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,barcode_clusters=barcode_sample_dict['WF40_ND21-251_HSC_3k'],rev=True)

    # Reimport adata objects (automatically named after the GTF/GFF in the analysis directory)
    #nd251_1_adata = ad.read_h5ad('ND251-1-isoform.h5ad')
    #nd251_2_adata = ad.read_h5ad('ND251-2-isoform.h5ad')
    # Ensure each has adata has capture specific cell barcodes and combine
    nd251_1_adata = iso.append_sample_name(nd251_1_adata, 'WF40_ND21-251_HSC_3k-1')
    nd251_2_adata = iso.append_sample_name(nd251_2_adata, 'WF40_ND21-251_HSC_3k-2')
    nd251_adata = ad.concat([nd251_1_adata, nd251_2_adata], axis=0, join='outer')
    nd251_adata.write_h5ad('ND251-apharesis-isoform.h5ad')
    pseudo_pdf, tpm, isoform_to_gene_ratio = iso.pseudo_cluster_counts(nd251_adata,cell_threshold=5,count_threshold=0,compute_tpm=True)
    pseudo_pdf.to_csv('nd251-combined_pseudo_cluster_counts.txt', sep='\t')
    tpm.to_csv('nd251-combined_pseudo_cluster-tpm.txt', sep='\t')
    isoform_to_gene_ratio.to_csv('nd251-combined_pseudo_cluster-ratio.txt', sep='\t')

    ######## Process data from elderly Bone Marrow samples ########
    # Import the cellHarmony results and save cell type pd frame for each cell barcode for each separate capture
    barcode_cluster_dir = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/Illumina/SoupX-hg38/cellHarmony/QueryGroups.cellHarmony.txt'
    barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dir)

    # Produce an junction counts h5ad for the AM72-N-0206_CD34
    matrix_path = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/AM72/scisoseq.seurat_info/isoforms_seurat'
    gff_source = 'AM72'
    am72_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,barcode_clusters=barcode_sample_dict['AM72-N-0206_CD34'],rev=True)
    #am72_adata = ad.read_h5ad('AM72-isoform-isoform.h5ad')
    pseudo_pdf, tpm, isoform_to_gene_ratio = iso.pseudo_cluster_counts(am72_adata,cell_threshold=5,count_threshold=0,compute_tpm=True)
    pseudo_pdf.to_csv('AM72-combined_pseudo_cluster_counts.txt', sep='\t')
    tpm.to_csv('AM72-combined_pseudo_cluster-tpm.txt', sep='\t')
    isoform_to_gene_ratio.to_csv('AM72-combined_pseudo_cluster-ratio.txt', sep='\t')

    # Produce an junction counts h5ad for the WF83-N-0215_CD34 capture
    matrix_path = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/WF83/scisoseq.seurat_info/isoforms_seurat'
    gff_source = 'WF83'
    wf83_adata = isor.exportConsensusIsoformMatrix(matrix_path,isoform_association_path,gff_source=gff_source,barcode_clusters=barcode_sample_dict['WF83-N-0215_CD34'],rev=True)
    #wf83_adata = ad.read_h5ad('WF83-isoform-isoform.h5ad')
    pseudo_pdf, tpm, isoform_to_gene_ratio = iso.pseudo_cluster_counts(wf83_adata,cell_threshold=5,count_threshold=0,compute_tpm=True)
    pseudo_pdf.to_csv('WF83-combined_pseudo_cluster_counts.txt', sep='\t')
    tpm.to_csv('WF83-combined_pseudo_cluster-tpm.txt', sep='\t')
    isoform_to_gene_ratio.to_csv('WF83-combined_pseudo_cluster-ratio.txt', sep='\t')



    #### Run AltAnalyze2
    #AltAnalyze.py --accessoryAnalysis mergeFiles --input /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/Isoforms/nd167-combined_pseudo_cluster-ratio.txt --input /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/Isoforms/nd251-combined_pseudo_cluster-ratio.txt --input /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/Isoforms/AM72-combined_pseudo_cluster-ratio.txt --input /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/Isoforms/WF83-combined_pseudo_cluster-ratio.txt --replaceAsZero False --join union --output /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/Isoforms
    #python AltAnalyze.py --species Hs --platform "3'array" --expdir /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/Isoforms/AltAnalyze/ExpressionInput/exp.ND-isoforms-sparse.txt  --groupdir /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/Isoforms/AltAnalyze/ExpressionInput/groups.ND-isoforms-sparse.txt --compdir /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/Isoforms/AltAnalyze/ExpressionInput/comps.ND-isoforms-sparse.txt --output /Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/pseudocounts/Isoforms/AltAnalyze --vendor Ensembl






integrate_healthy_malignant_junction = False
if integrate_healthy_malignant_junction:
    isoform_association_path = "/Users/saljh8/Dropbox/Revio/BoneMarrow/gff-output/isoform_links.txt"

    ################################### JUNCTION LEVEL ANALYSIS ###################################
    ######## Process data from sample ND167 ########
    # Import the cellHarmony results and save cell type pd frame for each cell barcode for each separate capture
    barcode_cluster_dir = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/QueryGroups.cellHarmony.txt'
    barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dir)
    
    # Produce an junction counts h5ad for the WM35_ND20-167__HSC_3k capture
    matrix_path = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/HSC_3k/filtered_matrices'
    gff1 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/HSC_3k/ND167_HSC-3k.gtf'
    hsc_nd167_adata = junc.exportJunctionMatrix(matrix_path, ensembl_exon_dir, gff1, barcode_clusters=barcode_sample_dict['WM35_ND20-167__HSC_3k'])

    # Produce an junction counts h5ad for the WM35_ND20-167__MPP_3k capture
    matrix_path = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/MPP_3k/filtered_matrices'
    gff2 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/MPP_3k/ND167_MPP_3k.gtf'
    mpp_nd167_adata = junc.exportJunctionMatrix(matrix_path, ensembl_exon_dir, gff2, barcode_clusters=barcode_sample_dict['WM35_ND20-167__MPP_3k'])

    # Produce an junction counts h5ad for the WM35_ND20-167__LMPP_3k capture
    matrix_path = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/LMPP_3k/filtered_matrices'
    gff3 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/LMPP_3k/ND167_LMPP_3k.gtf'
    lmpp_nd167_adata = junc.exportJunctionMatrix(matrix_path, ensembl_exon_dir, gff3, barcode_clusters=barcode_sample_dict['WM35_ND20-167__LMPP_3k'])

    # Reimport adata objects (automatically named after the GTF/GFF in the analysis directory)
    hsc_nd167_adata = ad.read_h5ad('ND167_HSC-3k.h5ad')
    mpp_nd167_adata = ad.read_h5ad('ND167_MPP_3k.h5ad')
    lmpp_nd167_adata = ad.read_h5ad('ND167_LMPP_3k.h5ad')

    # Ensure each has adata has capture specific cell barcodes and combine
    hsc_nd167_adata = iso.append_sample_name(hsc_nd167_adata, 'WM35_ND20-167__HSC_3k')
    mpp_nd167_adata = iso.append_sample_name(mpp_nd167_adata, 'WM35_ND20-167__MPP_3k')
    lmpp_nd167_adata = iso.append_sample_name(lmpp_nd167_adata, 'WM35_ND20-167__LMPP_3k')

    nd167_adata = ad.concat([hsc_nd167_adata, mpp_nd167_adata, lmpp_nd167_adata], axis=0, join='outer')
    nd167_adata.write_h5ad('ND167.h5ad')

    nd167_adata = ad.read_h5ad('ND167.h5ad')
    pseudo_pdf = iso.pseudo_cluster_counts(nd167_adata,cell_threshold=5,count_threshold=0)
    pseudo_pdf.to_csv('nd167-combined_pseudo_cluster_counts.txt', sep='\t')


    ######## Process data from sample ND251 ########

    # Import the cellHarmony results and save cell type pd frame for each cell barcode for each separate capture
    barcode_cluster_dir = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/Illumina/SoupX-hg38/cellHarmony/QueryGroups.cellHarmony.txt'
    barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dir)

    # Produce an junction counts h5ad for the WF40_ND21-251_HSC_3k capture (first run)
    matrix_path = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/PacBio/first_run/scisoseq.seurat_info/isoforms_seurat'
    gff4 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/PacBio/first_run/ND251-1.gff'
    nd251_1_adata = junc.exportJunctionMatrix(matrix_path, ensembl_exon_dir, gff4, barcode_clusters=barcode_sample_dict['WF40_ND21-251_HSC_3k'],rev=True)

    # Produce an junction counts h5ad for the WF40_ND21-251_HSC_3k capture (second run)
    matrix_path = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/PacBio/size_selected/scisoseq.seurat_info/isoforms_seurat'
    gff5 = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND251_MAS-Seq/PacBio/size_selected/ND251-2.gff'
    nd251_2_adata = junc.exportJunctionMatrix(matrix_path, ensembl_exon_dir, gff5, barcode_clusters=barcode_sample_dict['WF40_ND21-251_HSC_3k'],rev=True)

    # Reimport adata objects (automatically named after the GTF/GFF in the analysis directory)
    nd251_1_adata = ad.read_h5ad('ND251-1.h5ad')
    nd251_2_adata = ad.read_h5ad('ND251-2.h5ad')

    # Ensure each has adata has capture specific cell barcodes and combine
    nd251_1_adata = iso.append_sample_name(nd251_1_adata, 'WF40_ND21-251_HSC_3k-1')
    nd251_2_adata = iso.append_sample_name(nd251_2_adata, 'WF40_ND21-251_HSC_3k-2')

    nd251_adata = ad.concat([nd251_1_adata, nd251_2_adata], axis=0, join='outer')
    nd251_adata.write_h5ad('ND251-apharesis.h5ad')

    pseudo_pdf = iso.pseudo_cluster_counts(nd251_adata,cell_threshold=5,count_threshold=0)
    pseudo_pdf.to_csv('nd251-combined_pseudo_cluster_counts.txt', sep='\t')

    ######## Process data from elderly Bone Marrow samples ########
    # Import the cellHarmony results and save cell type pd frame for each cell barcode for each separate capture
    barcode_cluster_dir = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/Illumina/SoupX-hg38/cellHarmony/QueryGroups.cellHarmony.txt'
    barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dir)

    # Produce an junction counts h5ad for the AM72-N-0206_CD34
    matrix_path = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/AM72/scisoseq.seurat_info/isoforms_seurat'
    gff6 = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/AM72/AM72.gff'
    am72_adata = junc.exportJunctionMatrix(matrix_path, ensembl_exon_dir, gff6, barcode_clusters=barcode_sample_dict['AM72-N-0206_CD34'],rev=True)
    am72_adata = ad.read_h5ad('AM72.h5ad')
    pseudo_pdf = iso.pseudo_cluster_counts(am72_adata,cell_threshold=5,count_threshold=0)
    pseudo_pdf.to_csv('am72-combined_pseudo_cluster_counts.txt', sep='\t')

    # Produce an junction counts h5ad for the WF83-N-0215_CD34 capture
    matrix_path = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/WF83/scisoseq.seurat_info/isoforms_seurat'
    gff7 = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/PacBio/WF83/WF83.gff'
    wf83_adata = junc.exportJunctionMatrix(matrix_path, ensembl_exon_dir, gff7, barcode_clusters=barcode_sample_dict['WF83-N-0215_CD34'],rev=True)
    wf83_adata = ad.read_h5ad('WF83.h5ad')

    pseudo_pdf = iso.pseudo_cluster_counts(wf83_adata,cell_threshold=5,count_threshold=0)
    pseudo_pdf.to_csv('wf83-combined_pseudo_cluster_counts.txt', sep='\t')


    #### AML and MDS

    gff_a1 = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML/run1/5801-AML-1.gff'
    gff_a2 ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML/run2/5801-AML-2.gff'
    gff_a3 ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML-post/5801-9_341.gff'
    gff_a4 ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML-therapy/5801IR.gff'
    gff_a5 ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/MDS-post/5801-post.gff'
    gff_a6 ='/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/MDS-pre/5801-pre.gff'
    gff_a7 ='/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/on/Y1607.gff'
    gff_a8 ='/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/pre/Y1258.gff'
    gff_a9 ='/Users/saljh8/Dropbox/Revio/MDS-Revio/WF78-MDS/post/Y1737.gff'
    gff_a10 ='/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/post/Y5116.gff'
    gff_a11 ='/Users/saljh8/Dropbox/Revio/MDS-Revio/WF78-MDS/pre/Y1493.gff'
    gff_a12 ='/Users/saljh8/Dropbox/Revio/AML-Revio/AML7/run1/AML7.gff'
    gff_a13 ='/Users/saljh8/Dropbox/Revio/AML-Revio/AML7/run2/AML7-2.gff'
    gff_a14 ='/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/12-run1/AML12.gff'
    gff_a15 ='/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/12-run1/AML12.gff'
    gff_a16 ='/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/12-run2/AML12-2.gff'
    gff_a17 ='/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/13-run1/AML13.gff'
    gff_a18 ='/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/13-run2/AML13-2.gff'
    gff_a19 ='/Users/saljh8/Dropbox/Revio/AML-Revio/AML14/run1/AML14.gff'
    gff_a20 ='/Users/saljh8/Dropbox/Revio/AML-Revio/AML14/run2/AML14-2.gff'

    barcode_cluster_dir = '/Users/saljh8/Dropbox/Revio/Aged-Healthy-Revio/Illumina/SoupX-hg38/cellHarmony/QueryGroups.cellHarmony.txt'
    barcode_sample_dict = iso.import_barcode_clusters(barcode_cluster_dir)

    # Produce an junction counts h5ad for the 5801 captures
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML/run1/isoforms_seurat'
    p5801_aml_adata_1 = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_a1,rev=True,barcode_clusters=barcode_sample_dict['MDS5801-preVEN_342'])
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML/run2/isoforms_seurat'
    p5801_aml_adata_2 = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_a2,rev=True,barcode_clusters=barcode_sample_dict['MDS5801-preVEN_342'])

    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML-therapy/isoforms_seurat'
    p5801_aml_therapy_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_a4,rev=True,sample_id='5801')

    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/AML-post/isoforms_seurat'
    p5801_aml_post_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_a3,rev=True,sample_id='5801')

    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/MDS-post/isoforms_seurat'
    p5801_mds_post_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_a5,rev=True,sample_id='5801')

    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/5801-MDS/MDS-pre/isoforms_seurat'
    p5801_mds_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_a6,rev=True,sample_id='5801')

    p5801_aml_adata_1 = iso.append_sample_name(p5801_aml_adata_1, '5801-AML-1')
    p5801_aml_adata_2 = iso.append_sample_name(p5801_aml_adata_2, '5801-AML-2')
    p5801_aml_therapy_adata = iso.append_sample_name(p5801_aml_therapy_adata, '5801IR')
    p5801_aml_post_adata = iso.append_sample_name(p5801_aml_post_adata, '5801-9_341')
    p5801_mds_post_adata = iso.append_sample_name(p5801_mds_post_adata, '5801-post')
    p5801_mds_adata = iso.append_sample_name(p5801_mds_adata, '5801-pre')
    p5801_adata = ad.concat([p5801_aml_adata_1, p5801_aml_adata_2,p5801_aml_therapy_adata,p5801_aml_post_adata,p5801_mds_post_adata,p5801_mds_adata], axis=0, join='outer')
    p5801_adata.write_h5ad('5801_junction.h5ad')

    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/post/isoforms_seurat'
    gff_source = 'Y5116'
    am75_post_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='AM75')
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/pre/isoforms_seurat'
    gff_source = 'Y1258'
    am75_pre_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='AM75')
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/AM75-MDS/on/isoforms_seurat'
    gff_source = 'Y1607'
    am75_on_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='AM75')
    am75_post_adata = iso.append_sample_name(am75_post_adata, 'Y5116')
    am75_pre_adata = iso.append_sample_name(am75_pre_adata, 'Y1258')
    am75_on_adata = iso.append_sample_name(am75_on_adata, 'Y1607')
    am75_adata = ad.concat([am75_post_adata, am75_pre_adata, am75_on_adata], axis=0, join='outer')
    am75_adata.write_h5ad('am75.h5ad')

    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/WF78-MDS/post/isoforms_seurat'
    gff_source = 'Y1737'
    wf78_post_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='WF78')
    matrix_path = '/Users/saljh8/Dropbox/Revio/MDS-Revio/WF78-MDS/pre/isoforms_seurat'
    gff_source = 'Y1493'
    wf78_pre_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='WF78')
    wf78_post_adata = iso.append_sample_name(wf78_post_adata, 'Y1737')
    wf78_pre_adata = iso.append_sample_name(wf78_pre_adata, 'Y1493')
    wf78_adata = ad.concat([wf78_post_adata, wf78_pre_adata], axis=0, join='outer')
    wf78_adata.write_h5ad('wf78.h5ad')

    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML14/run1/isoforms_seurat'
    gff_source = 'AML14'
    aml14_1_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='AML14')
    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML14/run2/isoforms_seurat'
    gff_source = 'AML14-2'
    aml14_2_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='AML14')
    aml14_1_adata = iso.append_sample_name(aml14_1_adata, 'AML14')
    aml14_2_adata = iso.append_sample_name(aml14_2_adata, 'AML14-2')
    aml14_adata = ad.concat([aml14_1_adata, aml14_2_adata], axis=0, join='outer')
    aml14_adata.write_h5ad('aml14.h5ad')

    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/12-run1/isoforms_seurat'
    gff_source = 'AML12'
    aml12_1_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='AML12')
    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/12-run2/isoforms_seurat'
    gff_source = 'AML12-2'
    aml12_2_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='AML12')
    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/13-run1/isoforms_seurat'
    gff_source = 'AML13'
    aml12_3_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='AML12')
    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML12/13-run2/isoforms_seurat'
    gff_source = 'AML13-2'
    aml12_4_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='AML12')
    aml12_1_adata = iso.append_sample_name(aml12_1_adata, 'AML12')
    aml12_2_adata = iso.append_sample_name(aml12_2_adata, 'AML12-2')
    aml12_3_adata = iso.append_sample_name(aml12_3_adata, 'AML13')
    aml12_4_adata = iso.append_sample_name(aml12_4_adata, 'AML13-2')
    aml12_adata = ad.concat([aml12_1_adata, aml12_2_adata,aml12_3_adata,aml12_4_adata], axis=0, join='outer')
    aml12_adata.write_h5ad('aml12.h5ad')

    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML7/run1/isoforms_seurat'
    gff_source = 'AML7'
    aml7_1_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='AML7')
    matrix_path = '/Users/saljh8/Dropbox/Revio/AML-Revio/AML7/run2/isoforms_seurat'
    gff_source = 'AML7-2'
    aml7_2_adata = junc.exportJunctionMatrix(matrix_path,ensembl_exon_dir,gff_source=gff_source,rev=True,sample_id='AML7')
    aml7_1_adata = iso.append_sample_name(aml7_1_adata, 'AML7')
    aml7_2_adata = iso.append_sample_name(aml7_2_adata, 'AML7-2')
    aml7_adata = ad.concat([aml7_1_adata, aml7_2_adata], axis=0, join='outer')
    aml7_adata.write_h5ad('aml7.h5ad')

    am75_pre_adata = ad.read_h5ad('Y1258-isoform.h5ad')
    am75_post_adata = ad.read_h5ad('Y5116-isoform.h5ad')
    am75_post_adata = iso.append_sample_name(am75_post_adata, 'Y5116')
    am75_pre_adata = iso.append_sample_name(am75_pre_adata, 'Y1258')
    am75_adata = ad.concat([am75_pre_adata, am75_post_adata], axis=0, join='outer')
    am75_adata.write_h5ad('am75-pre-post.h5ad')
    exportRatios(am75_adata,'am75-pre-post')

    
    p5801_aml_post_adata = ad.read_h5ad('5801-9_341-isoform.h5ad')
    p5801_mds_post_adata = ad.read_h5ad('5801-post-isoform.h5ad')
    p5801_mds_adata = ad.read_h5ad('5801-pre-isoform.h5ad')
    p5801_aml_therapy_adata = ad.read_h5ad('5801IR-isoform.h5ad')
    p5801_aml_adata_2 = ad.read_h5ad('5801-AML-2-isoform.h5ad')
    p5801_aml_adata_1 = ad.read_h5ad('5801-AML-1-isoform.h5ad')

    p5801_aml_adata_1 = iso.append_sample_name(p5801_aml_adata_1, '5801-AML-1')
    p5801_aml_adata_2 = iso.append_sample_name(p5801_aml_adata_2, '5801-AML-2')
    p5801_aml_therapy_adata = iso.append_sample_name(p5801_aml_therapy_adata, '5801IR')
    p5801_aml_post_adata = iso.append_sample_name(p5801_aml_post_adata, '5801-9_341')
    p5801_mds_post_adata = iso.append_sample_name(p5801_mds_post_adata, '5801-post')
    p5801_mds_adata = iso.append_sample_name(p5801_mds_adata, '5801-pre')
    p5801_adata = ad.concat([p5801_aml_therapy_adata,p5801_aml_post_adata], axis=0, join='outer')
    p5801_adata.write_h5ad('5801-AML-therapy.h5ad')
    exportRatios(p5801_adata,'5801-AML-therapy')

    nd167_adata = ad.read_h5ad('ND167-isoform.h5ad')
    nd251_adata = ad.read_h5ad('ND251-apharesis-isoform.h5ad')
    am72_adata = ad.read_h5ad('AM72-isoform.h5ad')
    wf83_adata = ad.read_h5ad('WF83-isoform.h5ad')
    p5801_adata = ad.read_h5ad('5801.h5ad')
    am75_adata = ad.read_h5ad('am75.h5ad')
    wf78_adata = ad.read_h5ad('wf78.h5ad')
    aml14_adata = ad.read_h5ad('aml14.h5ad')
    aml12_adata = ad.read_h5ad('aml12.h5ad')
    aml7_adata = ad.read_h5ad('aml12.h5ad')
    
    a_adata = ad.read_h5ad('5801-9_341-isoform.h5ad')
    b_adata = ad.read_h5ad('5801-post-isoform.h5ad')
    c_adata = ad.read_h5ad('5801-pre-isoform.h5ad')
    d_adata = ad.read_h5ad('5801IR-isoform.h5ad')
    e_adata = ad.read_h5ad('5801-AML-2-isoform.h5ad')
    f_adata = ad.read_h5ad('5801-AML-1-isoform.h5ad')

    exportRatios(a_adata,'5801-9_341')
    exportRatios(b_adata,'5801-post')
    exportRatios(c_adata,'5801-pre')
    exportRatios(d_adata,'5801IR')
    exportRatios(e_adata,'5801-AML-2')
    exportRatios(f_adata,'5801-AML-1')

    exportRatios(nd167_adata,'ND167')
    exportRatios(nd251_adata,'ND251')
    exportRatios(am72_adata,'AM72')
    exportRatios(wf83_adata,'WF83')
    exportRatios(p5801_adata,'5801')
    exportRatios(am75_adata,'am75')
    exportRatios(wf78_adata,'wf78')
    exportRatios(aml14_adata,'aml14')
    exportRatios(aml12_adata,'aml12')
    exportRatios(aml7_adata,'aml12')
    
    combined_adata = ad.concat([nd167_adata, nd251_adata, am72_adata, wf83_adata, p5801_adata, am75_adata, wf78_adata, aml14_adata, aml12_adata, aml7_adata], axis=0, join='outer')
    pseudo_pdf, tpm, isoform_to_gene_ratio = iso.pseudo_cluster_counts(combined_adata,cell_threshold=5,count_threshold=0,compute_tpm=True)
    pseudo_pdf.to_csv('ND-MDS-AML-combined_pseudo_cluster_counts.txt', sep='\t')
    tpm.to_csv('ND-MDS-AM-combined_pseudo_cluster-tpm.txt', sep='\t')
    isoform_to_gene_ratio.to_csv('ND-MDS-AM-combined_pseudo_cluster-ratio.txt', sep='\t')

old_young_junction_integrate = False
if old_young_junction_integrate:
    ################################### JUNCTION LEVEL ANALYSIS ###################################
    ######## Process data from sample ND167 ########
    

    start_time = time.time()
    print(("Finding top " + str(top_n_differential_events) + " differential splicing events per NMF cluster..."))
    diff_events, remove_clusters, results_all = find_top_differential_events(exp_file=full_psi_file, groups_file=groups_file,
                                                                             metadata=metadata,
                                                                             output_loc=os.path.join(cd, filename),
                                                                             min_group_size=min_group_size, dPSI=dPSI,
                                                                             dPSI_p_val=dPSI_p_val,
                                                                             min_differential_events=min_differential_events,
                                                                             top_n=top_n_differential_events)
    elapsed_time = (time.time() - start_time) / 60  # Convert seconds to minutes
    print("--- %.3f minutes ---" % elapsed_time)


import pathlib
from altanalyze3.utilities.helpers import get_md5_sum


DATA_FOLDER = pathlib.Path(__file__).resolve().parents[1].joinpath("data")

CONTROL_MD5_SUMS = {
    "gene_model_all.tsv": "2b8531420d7916b9da21102586a96015",
    "Cal27P5-1.bam": "6e418cb604ea82c90d4b3e832afb4997",
    "Cal27P5-1.bam.bai": "6f0d6aceabade33f91ea69e5f036125c",
    "Cal27P5-2.bam": "8ce0bc51544144fd9b9bb1c66c172a61",
    "Cal27P5-2.bam.bai": "28bb5431b5d3bd1d6b8dc7a30a128011",
    "Cal27P5-3.bam": "3d6fdef5e90c16deed6c36b418c597c4",
    "Cal27P5-3.bam.bai": "c92fbf32ffcf874979c41051f0d2e16c",
    "Cal27P5-1-copy.bam": "6e418cb604ea82c90d4b3e832afb4997",
    "Cal27P5_1_aggregated_counts.bed.gz": "faee5946f7d18064773fc67e01e4ce5b",
    "Cal27P5_1_intcounts.bed": "6f5250f1a841438fadcbe1419966f133",
    "Cal27P5_1_juncounts.bed": "24f6a6ac5f831c7709102c235db03ec5",
    "Cal27P5_2_intcounts.bed": "2d9889a6520f96012ba7d4be5cdcfea3",
    "Cal27P5_2_juncounts.bed": "0ab1f2283ba349f9b08dc76d1f29dfe9",
    "Cal27P5_3_intcounts.bed": "be222bf62519f5b6c601377955e4ef76",
    "Cal27P5_3_juncounts.bed": "91863b9f8ce7c7d8fc2177b50ef0175e"
}


def test_all_md5_sums():
    collected_md5_sums = {}
    for location in DATA_FOLDER.glob("**/*"):
        if location.is_file():
            collected_md5_sums[location.name] = get_md5_sum(location)
    for filename, control_md5_sum in CONTROL_MD5_SUMS.items():
        assert filename in collected_md5_sums, "missing file"
        assert collected_md5_sums[filename] == control_md5_sum, "wrong MD5 sum"


import os
import sys
import csv
import anndata as ad
import pandas as pd
import numpy as np
from scipy.sparse import csr_matrix
from scipy.io import mmread
import collections
sys.path.insert(1, os.path.join(sys.path[0], '..'))
import long_read.isoform_junctions as junc
import long_read.isoform_matrix as iso
import long_read.isoform_ratios as isor
import long_read.isoform_translation as isot
import long_read.gff_process as gff_process
from Bio import SeqIO

query_gff_file = '/Users/saljh8/Dropbox/LongRead/PacBio/Ovarian/correct_merge.collapsed.gff'
ensembl_exon_dir = '/Users/saljh8/Documents/GitHub/altanalyze/AltDatabase/EnsMart91/ensembl/Hs/Hs_Ensembl_exon.txt'
#gff_output_dir = gff_process.consolidateLongReadGFFs(query_gff_file, ensembl_exon_dir)

ref_gff_file = "/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory/AltAnalyze3/gencode.v45.annotation.gff3"
genome_fasta = "/Users/saljh8/Dropbox/Revio/Other/Variants/SNV/genome.fa"
transcript_associations_file = "/Users/saljh8/Dropbox/LongRead/PacBio/Ovarian/gff-output/transcript_associations.txt"
cds_records, transcript_records, protein_records = isot.gff_translate(query_gff_file,genome_fasta,ref_gff_file,transcript_associations_file)

with open("transcript_sequences.fasta", "w") as cds_file:
    SeqIO.write(transcript_records, cds_file, "fasta")
    
with open("protein_sequences.fasta", "w") as protein_file:
    SeqIO.write(protein_records, protein_file, "fasta")

with open("cds_sequences.fasta", "w") as cds_file:
    SeqIO.write(cds_records, cds_file, "fasta")



import os, sys, csv, copy
import pandas as pd
sys.path.insert(1, os.path.join(sys.path[0], '..'))
from . import isoform_automate as isoa
from ..annotation import junction_isoform as ji
from ..oncosplice import metadataAnalysis as ma
import numpy as np

def compare_one_cluster_to_many(condition,cluster_order,groups_file,psi_matrix,junction_coords_file,gene_symbol_file,dataType='junction'):
    # Restrict by a specific condition (e.g., aged, young)
    print(f'Performing differential {dataType} analysis for every cluster within a single condition...')
    transcript_assoc_file = 'gff-output/transcript_associations.txt'
    protein_summary_file = 'protein_summary.txt'
    if dataType == 'junction':
        stats_folder = 'dPSI-cluster-events/'
    if dataType == 'isoform':
        stats_folder = 'diff-cluster-isoform/'
    if dataType == 'isoform-ratio':
        stats_folder = 'diff-cluster-ratio/'

    for cluster in cluster_order:
        # Filter samples for the current cluster
        current_cluster_samples = groups_file[
            (groups_file.index.str.startswith(cluster)) &  # Match the specific cluster
            (groups_file['grp'].str.endswith(condition))  # Ensure it belongs to the queried condition
        ]

        # Filter other samples within the same condition but in different clusters
        other_samples = groups_file[
            (~groups_file.index.str.startswith(cluster)) &  # Exclude the current cluster
            (groups_file['grp'].str.endswith(condition))  # Ensure it belongs to the queried condition
        ]

        # Label the groups appropriately
        current_cluster_samples = current_cluster_samples.copy()  # Optional: Ensure it is a copy
        current_cluster_samples.loc[:, 'grp'] = cluster  # Label as the current cluster (e.g., 'HSC-1')
        other_samples = other_samples.copy()  # Optional: Make a copy to avoid warnings
        other_samples.loc[:, 'grp'] = 'Others'  # Label as 'Others'

        # Combine the two groups
        filtered_groups_file = pd.concat([current_cluster_samples, other_samples])
        # Proceed with analysis if both groups have samples
        if not current_cluster_samples.empty and not other_samples.empty:
            print(f"Performing analysis for cluster: {cluster}")
            run_metadataAnalysis(cluster,psi_matrix,cluster,'Others',filtered_groups_file,stats_folder,filter_condition=condition)

    print ('Annotating the test results by gene, isoform and protein...')
    ji.annotate(gene_symbol_file,transcript_assoc_file,junction_coords_file,protein_summary_file,stats_folder,dataType=dataType)

def compare_two_groups_per_cluster(condition1,condition2,cluster_order,groups_file,psi_matrix,junction_coords_file,gene_symbol_file,dataType='junction'):
    print (f'Performing differential {dataType} analysis between two user supplied conditions...')
    transcript_assoc_file = 'gff-output/transcript_associations.txt'
    protein_summary_file = 'protein_summary.txt'
    if dataType == 'junction':
        stats_folder = 'dPSI-covariate-events/'
    if dataType == 'isoform':
        stats_folder = 'diff-covariate-isoform/'
    if dataType == 'isoform-ratio':
        stats_folder = 'diff-covariate-ratio/'
        
    for cluster in cluster_order:
        filtered_groups_file = groups_file[
            (groups_file.index.str.startswith(cluster)) &  # Filter by cell type prefix
            (groups_file['grp'].str.endswith((condition1, condition2)))  # Keep only 'aged' and 'young' groups
        ]
        try: run_metadataAnalysis(cluster,psi_matrix,condition1,condition2,filtered_groups_file,stats_folder)
        except:
            print (f'Failed for {condition1} vs. {condition2} for {cluster}')
    print ('Annotating the test results by gene, isoform and protein...')
    ji.annotate(gene_symbol_file,transcript_assoc_file,junction_coords_file,protein_summary_file,stats_folder,dataType=dataType)

def run_metadataAnalysis(cluster,psi_matrix,condition1,condition2,filtered_groups_file,stats_folder,filter_condition=False):
    matching_uids = filtered_groups_file.index.intersection(psi_matrix.columns)
    # Subset the PSI matrix to only include the matching sample columns
    subset_psi_matrix = psi_matrix[matching_uids]
    print (f'Exporting {condition1} vs. {condition2} for {cluster}')
    diff_stats = ma.mwuCompute(subset_psi_matrix, filtered_groups_file, grpvar='grp', min_group_size=2)

    if filter_condition == False:
        stats_file = f'{stats_folder}/{condition1}-{condition2}-{cluster}_stats.txt'
    else:
        stats_file = f'{stats_folder}/{filter_condition}-{condition1}-{condition2}-{cluster}_stats.txt'
    os.makedirs(os.path.dirname(stats_file), exist_ok=True)    
    diff_stats.to_csv(stats_file, sep='\t', index=True)

def compute_differentials(sample_dict,conditions,cluster_order,gene_symbol_file,analyses=['junction','isoform','isoform-ratio']):
    analyzed_intial_conditions = []
    for pair in conditions:
        (condition1,condition2) = pair
        print (f'Analyzing {condition2} vs. {condition1} ')

        if 'junction' in analyses:
            psi_dir = 'psi_combined_pseudo_cluster_counts.txt'
            junction_coords_file = 'junction_combined_pseudo_cluster_counts-filtered.txt'

            psi_matrix = pd.read_csv(psi_dir, sep='\t', index_col=0)
            print ('PSI matrix imported')

            # Extract 'groups' for each UID from the sample_dict
            uid_to_group = isoa.get_sample_to_group(sample_dict,'junction')

            # Create a DataFrame for the group mapping using the PSI matrix columns
            groups_file = pd.DataFrame({
                'grp': [uid_to_group.get(col.split('.')[1], 'unknown') for col in psi_matrix.columns],  # Extract UID and map to group
                'uid': psi_matrix.columns
            }).set_index('uid')

            # Compute Differential Splicing between groups and cell types
            if condition1 not in analyzed_intial_conditions:
                compare_one_cluster_to_many(condition1,cluster_order,groups_file,psi_matrix,junction_coords_file,gene_symbol_file,dataType='junction')
            compare_two_groups_per_cluster(condition1,condition2,cluster_order,groups_file,psi_matrix,junction_coords_file,gene_symbol_file,dataType='junction')

        if 'isoform' in analyses:
            # Compute Differential Isoform abundance between groups and cell types
            isoform_dir = 'isoform_combined_pseudo_cluster_tpm-filtered.txt'
            isoform_matrix = pd.read_csv(isoform_dir, sep='\t', index_col=0)
            log2_iso_matrix = np.log2(isoform_matrix + 1)
            uid_to_group = isoa.get_sample_to_group(sample_dict,'isoform')
            groups_file = pd.DataFrame({
                'grp': [uid_to_group.get(col.split('.')[1], 'unknown') for col in isoform_matrix.columns],  # Extract UID and map to group
                'uid': isoform_matrix.columns
            }).set_index('uid')
            if condition1 not in analyzed_intial_conditions:
                compare_one_cluster_to_many(condition1,cluster_order,groups_file,log2_iso_matrix,None,gene_symbol_file,dataType='isoform')
            compare_two_groups_per_cluster(condition1,condition2,cluster_order,groups_file,log2_iso_matrix,None,gene_symbol_file,dataType='isoform')

        if 'isoform-ratio' in analyses:
            isoform_dir = 'isoform_combined_pseudo_cluster_ratio-filtered.txt'
            isoform_matrix = pd.read_csv(isoform_dir, sep='\t', index_col=0)
            uid_to_group = isoa.get_sample_to_group(sample_dict,'isoform')
            groups_file = pd.DataFrame({
                'grp': [uid_to_group.get(col.split('.')[1], 'unknown') for col in isoform_matrix.columns],  # Extract UID and map to group
                'uid': isoform_matrix.columns
            }).set_index('uid')
            if condition1 not in analyzed_intial_conditions:
                compare_one_cluster_to_many(condition1,cluster_order,groups_file,isoform_matrix,None,gene_symbol_file,dataType='isoform-ratio')
            compare_two_groups_per_cluster(condition1,condition2,cluster_order,groups_file,isoform_matrix,None,gene_symbol_file,dataType='isoform-ratio')
            analyzed_intial_conditions.append(condition1)



""" This module combines information across GFF files to compare isoform structures, 
annotate exons/transcripts and to nominate the longest exemplar isoforms. Note that
this script is biased against isoforms that are bleeding and internal hybrid."""

import os,sys
from collections import defaultdict
from tqdm import tqdm
import pandas as pd
import collections

novelGene = 0
novelGeneLoci = [1,1]
additional_junctions={}

def findNovelSpliceSite(geneID, position, strand):
    exons = geneData[geneID]
    for i, (start, stop, exon) in enumerate(exons):
        """
        if strand == '-' and position>258251:
            print (geneID, position, strand)
            print (exons,start,position,stop,exon)
            pass"""
        if (strand == '+' and position < start) or (strand == '-' and position > start):
            #print ('a');sys.exit()
            if i == 0:
                return f"U1.1_{position}" # Position is upstream of the first exon
            else:
                pass # not in the exon/intron
        elif (strand == '+' and start <= position <= stop) or (strand == '-' and stop <= position <= start):
            #print ('b');sys.exit()
            if 'I' in exon and exon.endswith('.1'): # Position is within the current intron
                #print ('c');sys.exit()
                if (strand == '+' and (position - start < 50)) or (strand == '-' and (start - position < 50)):
                    #if strand == '-':  print(1, exon,position,start, stop, strand,f"{exons[i - 1][2]}_{position}");sys.exit()
                    # Within 50nt of the beginning of the intron - alt SS of prior exon region
                    return f"{exons[i - 1][2]}_{position}"
                elif (strand == '+' and (stop - position < 50)) or (strand == '-' and (position - stop < 50)):
                    #print(1, exon,position,start, stop, strand,f"{exons[i - 1][2]}_{position}")
                    #print ('d');sys.exit()
                    # Within 50nt of the end of the intron - alt SS of next exon region
                    try:
                        return f"{exons[i + 1][2]}_{position}"
                    except:
                        print (position,exons);sys.exit()
                else: # In the middle of the intron
                    #print ('k');sys.exit()
                    return f"{exon}_{position}"
            else:
                #print ('e');sys.exit()
                # Position is within the current exon
                return f"{exon}_{position}"
    #print ('f');sys.exit()
    last_exon_id = exons[-1][2] # Position is after the last exon
    prefix, number = last_exon_id[0], last_exon_id[1:].split('.')
    next_number = str(int(number[0]) + 1)
    return f"U{next_number}.1_{position}"

def rangeCheck(exon_range,novelGeneLoci):
    if exon_range[0] >= novelGeneLoci[0] and exon_range[0] <= novelGeneLoci[1]:
        return True
    elif exon_range[1] >= novelGeneLoci[0] and exon_range[1] <= novelGeneLoci[1]:
        return True
    else:
        return False

def exonAnnotate(chr, exons, strand, transcript_id):
    """Assign exon/intron-level annotations"""
    global novelGene
    global novelGeneLoci
    global additional_junctions

    exon_range = [exons[0][0],exons[-1][-1]]
    ### First determine the genes and initial known assignments
    associations = []
    genes = []
    if strand == '-':
        if exon_range[1]>exon_range[0]: ### Gencode is ordered transcriptomically
            exons.reverse()
        exons = [(b, a) for a, b in exons]
    for (pos1, pos2) in exons:
        if (chr, pos1, strand, 1) in exonCoordinates:
            gene1, exon1, ind1 = exonCoordinates[chr, pos1, strand, 1]
            if gene1 not in genes:
                genes.append(gene1)
        else:
            gene1, exon1, strand1 = None, None, None
        if (chr, pos2, strand, 2) in exonCoordinates:
            gene2, exon2, ind2 = exonCoordinates[chr, pos2, strand, 2]
            if gene2 not in genes:
                genes.append(gene2)
        else:
            gene2, exon2, strand2 = None, None, None
        if gene1 is None and gene2 is None:
            association = 'novel', (pos1,pos2)
        elif gene1 is not None and gene2 is not None:
            if gene1 == gene2:
                if exon1 == exon2:
                    association = 'common', (gene1, exon1)
                else:
                    association = 'mixed', (gene1, exon1, exon2)
            else:
                # Trans-splicing
                association = 'trans', (gene1, gene2, exon1, exon2)
        else:
            if gene1 is None:
                association = 'novel1', (gene2, exon2, pos1)
            else:
                association = 'novel2', (gene1, exon1, pos2)
        associations.append(association)

    ### Determine the spectrum of possible gene IDs (trans-splicing)
    if len(genes) == 0:
        """
        if 'PB.9.48' == transcript_id:
            print(novelGeneLoci,exon_range);sys.exit()
        """
        if rangeCheck(exon_range,novelGeneLoci) == False:
            novelGene += 1
        gene = 'UNK' + str(novelGene)
        novelGeneLoci = exon_range
    elif len(genes) > 1:
        if len(genes) == 3:
            #print('Multi-Trans-splicing', genes)
            pass
        gene = genes[0]
    else:
        gene = genes[0]
    current_gene = gene # used to document current gene for trans-splicing

    ### Annotate novel exon/intron IDs and include in-between regions in the final isoform annotation
    if len(genes) == 0:
        output_list = [f"{start}-{end}" for start, end in exons]
        associations_updated = output_list  # Novel genes don't require exon annotations (likely update - currently position tuple)
        associations_simple = output_list
    else:
        """
        if 'PB.2986.5' == transcript_id:
            print(exons,associations);sys.exit()
        """
        associations_updated = []
        associations_simple = []
        for ind, association in enumerate(associations):
            exonType = association[0]

            if exonType == 'common':
                gene1 = association[1][0]
                exonAnnotation = association[1][1]
                if gene1 != gene or current_gene != gene:
                    exonAnnotation = gene1 + ':' + exonAnnotation
                associations_updated.append(exonAnnotation) # denotes a splice site or end
                associations_simple.append(exonAnnotation)
            elif exonType == 'mixed':
                exonAnnotation1 = association[1][1]
                exonAnnotation2 = association[1][2]
                gene1 = association[1][0]
                start,stop = exons[ind]
                start_index = exonCoordinates[chr, start, strand, 1][-1]
                end_index = exonCoordinates[chr, stop, strand, 2][-1]+1
                if gene1 != gene or current_gene != gene:
                    exonAnnotation1 = gene1 + ':' + exonAnnotation1
                    current_gene = gene1
                associations_updated.append(exonAnnotation1)
                if ind != 0: # don't add the begining of the exon
                    associations_simple.append(exonAnnotation1)
                prior_splice_acceptor = None
                ref_exons = geneData[current_gene]
                # Add exon and intron regions in the middle of a broader exon
                for in_between_region in ref_exons[start_index:end_index]:
                    if current_gene != gene: # trans-splicing
                        associations_updated.append(current_gene+':'+in_between_region[-1])
                    else:
                        associations_updated.append(in_between_region[-1])
                    current_spice_donor = in_between_region[0]
                    if prior_splice_acceptor != None:
                        if transcript_id not in additional_junctions:
                            additional_junctions[transcript_id]=[(prior_splice_acceptor,current_spice_donor)]
                        else:
                            additional_junctions[transcript_id].append((prior_splice_acceptor,current_spice_donor))
                    prior_splice_acceptor = in_between_region[-2]
                if gene1 != gene or current_gene != gene:
                    exonAnnotation2 = gene1 + ':' + exonAnnotation2
                    current_gene = gene1
                associations_updated.append(exonAnnotation2) # denotes a splice site or end
                if ind != (len(associations)-1): # Don't add the last splice site
                    associations_simple.append(exonAnnotation2)
                """
                if len(ref_exons[start_index:end_index])>1 and 'I2.1' in associations_updated and strand == '-':
                    print(exons)
                    print(associations)
                    print('****',start_index,end_index,ref_exons[start_index:end_index])
                    print(associations_updated);sys.exit()"""
            elif exonType == 'trans':
                gene1 = association[1][0]
                gene2 = association[1][1]
                exonAnnotation1 = association[1][2]
                exonAnnotation2 = association[1][3]
                if gene == gene1:
                    associations_updated.append(exonAnnotation1)
                    associations_simple.append(exonAnnotation1)
                    associations_updated.append(gene2 + ':' + exonAnnotation2)
                    associations_simple.append(gene2 + ':' + exonAnnotation2)
                elif gene == gene2:
                    associations_updated.append(gene1 + ':' + exonAnnotation1)
                    associations_simple.append(gene1 + ':' + exonAnnotation1)
                    associations_updated.append(exonAnnotation2) # denotes a splice site or end
                    associations_simple.append(exonAnnotation2)
                    current_gene = gene2

            elif exonType == 'novel1':
                novel_position = association[1][2]
                exonAnnotation2 = association[1][1]
                gene2 = association[1][0]
                hits=[]
                for geneID in genes: ### more than one possible gene the novel exon is in
                    #if len(genes)>1: 
                    #print(association,strand,gene2,novel_position,genes);sys.exit()
                    exonAnnotation1 = findNovelSpliceSite(geneID, novel_position, strand)
                    hits.append([exonAnnotation1,geneID])
                    #if len(genes)>1: print(exonAnnotation1,geneID)
                    #if 'U' not in exonAnnotation1:
                hits.sort() # E and I annotations will be prioritized
                exonAnnotation1,geneID = hits[0]
                """if strand == '-' and novel_position>164713:
                    print (exonAnnotation1);sys.exit()"""

                if gene != gene2 or current_gene != gene: # trans-splicing somewhere else in this transcript
                    exonAnnotation1 = gene2+':'+exonAnnotation1
                    current_gene = gene2
                associations_updated.append(exonAnnotation1)
                associations_simple.append(exonAnnotation1)
                # If exons and/or introns span from the begining of a novel exon to a known end, include all regions in between (e.g., U1.1_29934 to E3.1)
                # start_index will not exist in the database (novel end)
                start,stop = exons[ind]
                end_index = exonCoordinates[chr, stop, strand, 2][-1]
                evaluate_between_regions = True
                if 'U' in exonAnnotation1:
                    start_index = 0
                else:
                    try:
                        if 'I' in exonAnnotation1:
                            # If the splice site is in the intron the last portion of the intron is retained
                            # Thus, include the intron splice site coordinates in the search to denote IR
                            start,stop = exonData[geneID,exonAnnotation1.split("_")[0].replace('I','E')]
                            start_index = exonCoordinates[chr, start, strand, 1][-1] + 1
                        else:
                            start,stop = exonData[geneID,exonAnnotation1.split("_")[0]]
                            start_index = exonCoordinates[chr, start, strand, 1][-1]
                    except:
                        evaluate_between_regions = False

                if evaluate_between_regions:
                    ref_exons = geneData[current_gene]
                    prior_splice_acceptor = None
                    # Add exon and intron regions in the middle of a broader exon
                    for in_between_region in ref_exons[start_index:end_index]:
                        if current_gene != gene: # trans-splicing
                            associations_updated.append(current_gene+':'+in_between_region[-1])
                        else:
                            associations_updated.append(in_between_region[-1])
                        current_spice_donor = in_between_region[0]
                        if prior_splice_acceptor != None:
                            if transcript_id not in additional_junctions:
                                additional_junctions[transcript_id]=[(prior_splice_acceptor,current_spice_donor)]
                            else:
                                additional_junctions[transcript_id].append((prior_splice_acceptor,current_spice_donor))
                        prior_splice_acceptor = in_between_region[-2]

                if gene != gene2 or current_gene != gene: # trans-splicing somewhere else in this transcript
                    exonAnnotation2 = gene2+':'+exonAnnotation2
                    current_gene = gene2
                associations_updated.append(exonAnnotation2) # denotes a splice site or end
                associations_simple.append(exonAnnotation2)
            elif exonType == 'novel2':
                novel_position = association[1][2]
                exonAnnotation1 = association[1][1]
                gene1 = association[1][0]
                hits=[]
                for geneID in genes: ### more than one possible gene the novel exon is in
                    exonAnnotation2 = findNovelSpliceSite(geneID, novel_position, strand)
                    hits.append([exonAnnotation2,geneID])
                    #if len(genes)>1: print(exonAnnotation1,geneID)
                    #if 'U' not in exonAnnotation1:
                hits.sort() # E and I annotations will be prioritized
                exonAnnotation2,geneID = hits[0]
                """if strand == '-' and novel_position>258251:
                    print (exonAnnotation2);sys.exit()"""

                if gene != gene1 or current_gene != gene: # trans-splicing somewhere else in this transcript
                    exonAnnotation1 = gene1+':'+exonAnnotation1
                    current_gene = gene1
                associations_updated.append(exonAnnotation1)
                associations_simple.append(exonAnnotation1)
                # If exons and/or introns span from the begining of a known exon to the novel end, include all regions in between (e.g., E10.1 to U15.1_1234)
                # end_index will not exist in the database (novel end)
                start,stop = exons[ind]
                start_index = exonCoordinates[chr, start, strand, 1][-1]
                evaluate_between_regions = True
                if 'U' in exonAnnotation2:
                    end_index = None
                else:
                    try:
                        eid = exonAnnotation2.split("_")[0].replace('I','E')
                        start,stop = exonData[geneID,eid]
                        end_index = exonCoordinates[chr, stop, strand, 2][-1]
                        if 'I' in exonAnnotation2 and start_index == end_index: # (E3.1, I3.1_12345) - include the exon-intron boundary (bleeding exon)
                            prefix, number = eid[0], eid[1:].split('.')
                            eid2 =  f"E{str(int(number[0]) + 1)}.1"
                            start,stop = exonData[geneID,eid2]
                            end_index = exonCoordinates[chr, stop, strand, 2][-1]
                    except:
                        # Intronic position is associated with a strange region, such as ENSG00000127483-I15.2
                        evaluate_between_regions = False
                if evaluate_between_regions:
                    ref_exons = geneData[current_gene]
                    prior_splice_acceptor = None
                    # Add exon and intron regions in the middle of a broader exon
                    for in_between_region in ref_exons[start_index:end_index]:
                        if current_gene != gene: # trans-splicing
                            associations_updated.append(current_gene+':'+in_between_region[-1])
                        else:
                            associations_updated.append(in_between_region[-1])
                        current_spice_donor = in_between_region[0]
                        if prior_splice_acceptor != None:
                            if transcript_id not in additional_junctions:
                                additional_junctions[transcript_id]=[(prior_splice_acceptor,current_spice_donor)]
                            else:
                                additional_junctions[transcript_id].append((prior_splice_acceptor,current_spice_donor))
                        prior_splice_acceptor = in_between_region[-2]

                if gene != geneID or current_gene != gene: 
                    exonAnnotation2 = geneID+':'+exonAnnotation2
                    current_gene = geneID
                associations_updated.append(exonAnnotation2) # denotes a splice site or end
                associations_simple.append(exonAnnotation2)

            elif exonType == 'novel':
                novel_position1 = association[1][0]
                novel_position2 = association[1][1]
                """
                hits=[]
                for geneID in genes: ### more than one possible gene the novel exon is in
                    exonAnnotation1 = findNovelSpliceSite(geneID, novel_position1, strand)
                    hits.append([exonAnnotation1,geneID])
                hits.sort() # E and I annotations will be prioritized
                exonAnnotation1,geneID = hits[0]
                """
                exonAnnotation1 = findNovelSpliceSite(current_gene, novel_position1, strand)
                if current_gene != gene: 
                    exonAnnotation1 = current_gene+':'+exonAnnotation1
                """
                hits=[]
                for geneID in genes: ### more than one possible gene the novel exon is in
                    exonAnnotation2 = findNovelSpliceSite(geneID, novel_position2, strand)
                    hits.append([exonAnnotation2,geneID])
                hits.sort() # E and I annotations will be prioritized
                exonAnnotation2,geneID = hits[0]
                """
                exonAnnotation2 = findNovelSpliceSite(current_gene, novel_position2, strand)
                if current_gene != gene: 
                    exonAnnotation2 = current_gene+':'+exonAnnotation2
                ### Need to fill in-between exons
                associations_updated.append(exonAnnotation1)
                associations_updated.append(exonAnnotation2) # denotes a splice site or end
                associations_simple.append(exonAnnotation1)
                associations_simple.append(exonAnnotation2)
    """
    if 'PB.2986.5' == transcript_id:
        print('***',associations)
        print(associations_updated);sys.exit()
    """
    associations_updated = uniqueList(associations_updated)
    associations_simple = uniqueList(associations_simple)

    return gene,associations_updated,associations_simple,genes

def uniqueList(ls):
    return [seen.add(x) or x for seen in (set(),) for x in ls if x not in seen]

def knownIsoform(transcripts):
    search_strings = ["ENST", "NM_", "XM_", "XR_"]
    found = any(any(s in transcript for s in search_strings) for transcript in transcripts)
    if found:
        return 'known'
    else:
        return 'novel'

def selectKnownIsoform(transcripts):
    search_strings = ["ENST", "NM_", "XM_", "XR_"]
    found=[]
    for transcript in transcripts:
        for s in search_strings:
            if transcript.startswith(s):
                found.append(transcript)
    found.sort()
    for transcript in transcripts:
        if transcript not in found:
            found.append(transcript)
    return found
    
def collapseIsoforms(gene_db, junction_db, gff_organization, mode):
    num_isoforms = 0
    collaped_db = defaultdict(lambda: defaultdict(list))
    
    # Adding a progress bar for the outer loop iterating over genes
    for gene, isoforms in tqdm(gene_db.items(), desc="Collapsing Isoforms"):
        sorted_isoforms = sorted(isoforms, key=len, reverse=True)
        # Initially set every isoform as a key with an empty list
        for isoform in sorted_isoforms:
            collaped_db[gene][isoform] = []
        # Populate the list with sub-isoforms
        for i, isoform in enumerate(sorted_isoforms):
            for shorter_isoform in sorted_isoforms[i+1:]:
                # Check if one isoform is a subset of another
                if shorter_isoform in isoform:
                    if isoform != shorter_isoform:
                        collaped_db[gene][isoform].append(shorter_isoform)
        # Remove super-isoforms that are present as sub-isoforms
        for super_isoform in list(collaped_db[gene].keys()):
            for sub_isoforms in collaped_db[gene].values():
                if super_isoform in sub_isoforms:
                    del collaped_db[gene][super_isoform]
                    break

        num_isoforms += len(collaped_db[gene])
    print(num_isoforms, 'collapsed unique isoforms')

    if mode == 'Ensembl':
        def get_transcript_ids(gene, isoforms):
            transcript_ids = defaultdict(lambda: defaultdict(list))
            for iso in isoforms:
                for (file, info) in junction_db[gene, iso]:
                    transcript_id = info.split(';')[gff_organization[file][0]].split(gff_organization[file][1])[1]
                    transcript_ids[transcript_id] = iso
            return transcript_ids

        num_isoforms = 0

        super_isoform_collapsed_db = collaped_db
        collaped_db = defaultdict(lambda: defaultdict(list))
        keys_added = {}

        # Adding progress bar for Ensembl processing
        for (gene, isoform) in tqdm(junction_db, desc="Processing Ensembl Isoforms"):
            super_isoforms = super_isoform_collapsed_db[gene]
            if isoform in super_isoforms:  # Hence, super_isoform - keep
                sub_isoforms = super_isoform_collapsed_db[gene][isoform]
                combined_iso = [isoform] + sub_isoforms
                transcript_ids = get_transcript_ids(gene, combined_iso)
                transcript_list = list(transcript_ids.keys())
                found = next((x for x in transcript_list if x.startswith('ENST')), False)
                if found:
                    ens_iso = transcript_ids[found]
                    if ens_iso == isoform:
                        collaped_db[gene][ens_iso] = []
                    else:
                        collaped_db[gene][ens_iso].append(isoform)
                    keys_added[gene, isoform] = []
                    keys_added[gene, ens_iso] = []
            else:
                for iso in super_isoforms:
                    if isoform in super_isoforms[iso]:
                        combined_iso = [iso] + super_isoforms[iso]
                        transcript_ids = get_transcript_ids(gene, combined_iso)
                        transcript_list = list(transcript_ids.keys())
                        found = next((x for x in transcript_list if x.startswith('ENST')), False)
                        if found:
                            ens_iso = transcript_ids[found]
                            if ens_iso == iso:
                                collaped_db[gene][ens_iso] = []
                            else:
                                collaped_db[gene][ens_iso].append(isoform)
                            keys_added[gene, isoform] = []
                            keys_added[gene, ens_iso] = []

        # Add isoforms that were not processed earlier
        for (gene, isoform) in junction_db:
            if (gene, isoform) not in keys_added:
                collaped_db[gene][isoform] = []

        # Count the total number of collapsed isoforms
        for gene in collaped_db:
            num_isoforms += len(collaped_db[gene])
        
        print(num_isoforms, 'collapsed Ensembl prioritized unique isoforms')

    return collaped_db



def consolidateLongReadGFFs(directory, exon_reference_dir, mode="collapse"):
    """Only return isoforms with unique splice-site combinations"""
    junction_db = collections.OrderedDict()
    junction_str_db = collections.OrderedDict()
    gene_db = collections.OrderedDict()
    strand_db = collections.OrderedDict()
    trans_spliced_isoforms = collections.OrderedDict()
    global novelGene
    global novelGeneLoci
    global additional_junctions

    # Import preprocessed Ensembl exon information
    importEnsemblGenes(exon_reference_dir)
    
    collapse_isoforms = False
    if isinstance(directory, list):
        # list of gff's from different locations
        files = directory
        directory = os.getcwd() # present working directory
        if len(files)>1:
            collapse_isoforms = True
    elif '.g' in directory.lower():
        # Export exon structure information only
        gff = directory
        directory = os.path.dirname(gff)
        gff = os.path.basename(gff)
        files = [gff]
    else:
        # Combine and report collapsed unique isoform structures 
        files = [file for file in os.listdir(directory) if file.endswith('.gff')]
        files += [file for file in os.listdir(directory) if file.endswith('.gtf')]
        #files = [file for file in os.listdir(directory) if file.endswith('.txt')]
        if len(files)>1:
            collapse_isoforms = True
    
    combined_dir = os.path.join(directory, 'gff-output')
    if not os.path.exists(combined_dir):
        os.makedirs(combined_dir)
    transcript_associations = os.path.join(combined_dir, 'transcript_associations.txt')
    eo = open(transcript_associations, 'w')

    def getJunctions(exons):
        splice_junctions = []
        if len(exons) == 1:
            splice_junctions = exons  # intron retention or single-exon transcript
        elif strand == '+':
            for i in range(len(exons) - 1):
                splice_junctions.append((exons[i][1], exons[i + 1][0]))
        else:
            for i in range(len(exons) - 1, 0, -1):
                splice_junctions.append((exons[i][0], exons[i - 1][1]))
        return splice_junctions

    def storeJunctions(junctions):
        for coords in junctions:
            uid = chr+':'+str(coords[0])+'-'+str(coords[1])
            if uid in unique_junction_db:
                unique_junction_db[uid].append(file)
            else:
                unique_junction_db[uid] = [file]

    def exon_str(exonIDs):
        separator = '|'
        exonIDs_str = separator.join(exonIDs)
        filtered_exonIDs = exonIDs
        if '_' in filtered_exonIDs[0]:
            del filtered_exonIDs[0]
        if '_' in filtered_exonIDs[-1]:
            del filtered_exonIDs[-1]
        filtered_exonIDs_str = separator.join(filtered_exonIDs) + '|'
        return filtered_exonIDs_str,exonIDs_str

    def process_isoform(chr, strand, info, exons, file):
        transcript_id = info.split(';')[ti].split(td)[1]
        gene, exonIDs, exonIDs_simple, genes = exonAnnotate(chr, exons, strand, transcript_id)
        # Restrict the exon string to high confidence exon boundaries
        filtered_exonIDs_str,exonIDs_str = exon_str(list(exonIDs))
        try:
            if mode == 'collapse':
                filtered_exonIDs_str,exonIDs_simple_str = exon_str(list(exonIDs_simple))
        except:
            pass
        """
        if transcript_id == 'ENST00000525626.6' or transcript_id == 'ENST00000682936.1':
            print (filtered_exonIDs_str,transcript_id,mode) 
        """
        if len(genes)>2:
            trans_spliced_isoforms[exonIDs_str] = []
        eo.write('\t'.join([gene, strand, exonIDs_str, transcript_id, file]) + '\n')
        #splice_junctions = getJunctions(exons)
        if 'UNK' not in gene:
            #junction_str_db[gene,filtered_exonIDs_str] = tuple(splice_junctions)
            try: 
                junction_db[(gene, filtered_exonIDs_str)].append((file, info))
            except:
                junction_db[(gene, filtered_exonIDs_str)] = [(file, info)]
            try:
                if filtered_exonIDs_str not in gene_db[gene]:
                    gene_db[gene].append(filtered_exonIDs_str)
            except:
                    gene_db[gene] = [filtered_exonIDs_str]
            strand_db[gene] = strand

    gff_organization={}
    for file in files:
        if os.path.exists(file):
            fn = file
            file = os.path.basename(file)
        else:
            fn = os.path.join(directory, file)
        file = file.split('.g')[0]
        firstRow = True
        exons = []
        isoforms = 2
        with open(fn, 'r') as filepath:
            for line in filepath:
                if line[0] == '#':
                    continue
                data = line.strip()
                t = data.split('\t')
                try: chr, null, type, pos1, pos2, null, strand, null, info = t
                except: 
                    print (info)
                    continue
                pos1 = int(pos1)
                pos2 = int(pos2)
                if 'chr' not in chr:
                    chr = 'chr'+chr

                if firstRow:
                    # Determine if the transcript is listed first or second
                    if type != 'transcript' and type != 'exon':
                        continue
                    ti = next((i for i, x in enumerate(info.split(';')) if 'transcript_id' in x), -1)
                    if '=' in info:
                        td = '='
                    else:
                        td = '"'
                    gff_organization[file]=ti,td
                    firstRow = False
                else:
                    #print(file,type,chr,pos1,pos2,info);sys.exit()
                    if type == 'transcript':
                        isoforms += 1
                        chr, strand, info = gene_info
                        process_isoform(chr, strand, info, exons, file)
                        exons = []
                    elif type != 'transcript' and type != 'exon':
                        pass
                    else:
                        exons.append((pos1, pos2))
                        gene_info = chr, strand, info
        # for the last isoform in the file
        chr, strand, info = gene_info
        process_isoform(chr, strand, info, exons, file)
        print(file, '...', isoforms, 'isoforms')

    eo.close()
    print(len(junction_db), 'unique isoforms')

    def sort_isoforms_with_ENST_first(junction_db):
        sorted_junction_db = defaultdict(list)
        for key, value in junction_db.items():
            sorted_value = sorted(value, key=lambda x: 'ENST' not in x[1])
            sorted_junction_db[key] = sorted_value
        return sorted_junction_db

    junction_db = sort_isoforms_with_ENST_first(junction_db)

    if collapse_isoforms == False:
        return transcript_associations
    else:
        """ Most isoforms should be redundant between samples - collapse isoforms based on redundant junctions
        and export a combined minimal GFF and gff-specific associated isoforms to the longest exemplar """ 

        # test case for collapseIsoforms: should give
        # {((1, 2), (3, 4), (5, 6), (7, 8)): [((1, 2), (3, 4), (5, 6))], ((1, 2), (3, 4), (7, 8)): [((1, 2), (3, 4), (7, 8)), ((3, 4), (7, 8))]}
        #a = {'gene1':[((1,2),(3,4),(5,6)),((1,2),(3,4),(5,6),(7,8)),((1,2),(3,4),(7,8)),((3,4),(7,8)),((1,2),(3,4),(7,8))]}
        
        if mode == 'collapse' or mode == 'Ensembl':
            super_isoform_db = collapseIsoforms(gene_db,junction_db,gff_organization,mode)
        else:
            super_isoform_db = defaultdict(lambda: defaultdict(list))
            for (gene, isoform) in junction_db: 
                super_isoform_db[gene][isoform] = []
        
        def get_transcript_ids(gene,isoforms):
            transcript_ids=defaultdict(lambda: defaultdict(list))
            for iso in isoforms:
                for (file, info) in junction_db[gene, iso]:
                    transcript_id = info.split(';')[gff_organization[file][0]].split(gff_organization[file][1])[1]
                    transcript_ids[transcript_id]=file
            return transcript_ids

        # Export super- to sub-isoform associations:
        eo = open(os.path.join(combined_dir, 'isoform_links.txt'), 'w')
        #jo = open(os.path.join(combined_dir, 'isoform_junctions.txt'), 'w')
        ao = open(os.path.join(combined_dir, 'isoform_annotations.txt'), 'w')
        isoforms_to_retain = {}
        a=0
        total_genes = len(super_isoform_db)
        # Use tqdm to create a progress bar
        for gene in tqdm(super_isoform_db, total=total_genes, desc="Processing genes"):
            added = []
            for isoform in super_isoform_db[gene]: 
                # isoform is a tuple of junction coordinates tuples
                strand = strand_db[gene]
                related_transcripts = []
                if 'UNK' not in gene:
                    transcript_ids = get_transcript_ids(gene,[isoform])
                    transcript_list = list(transcript_ids.keys())
                    ordered_transcripts = selectKnownIsoform(transcript_list)
                    ref_super_transcript_id = ordered_transcripts[0]
                    ref_super_file = transcript_ids[ref_super_transcript_id]
                    isoforms_to_retain[(ref_super_file, ref_super_transcript_id)] = [] 
                    related_transcripts.append(ref_super_transcript_id)
                    if len(ordered_transcripts)==1:
                        eo.write('\t'.join([gene, ref_super_transcript_id, ref_super_file, '', '']) + '\n')
                        a += 1
                    else: 
                        for t in transcript_list:
                            if t!=ref_super_transcript_id:
                                t_file =  transcript_ids[t]
                                eo.write('\t'.join([gene, ref_super_transcript_id, ref_super_file, t, t_file]) + '\n')
                                related_transcripts.append(ref_super_file)
                    for sub_isoform in super_isoform_db[gene][isoform]:
                        for (f3, info) in junction_db[gene, sub_isoform]:
                            sub_transcript_id = info.split(';')[gff_organization[f3][0]].split(gff_organization[f3][1])[1]
                            if (sub_transcript_id, f3) not in added:
                                if ref_super_transcript_id !=sub_transcript_id:
                                    eo.write('\t'.join([gene, ref_super_transcript_id, file, sub_transcript_id, f3]) + '\n')
                                    added.append((sub_transcript_id, f3))
                                    related_transcripts.append(sub_transcript_id)
                                    a += 1
                    known_isoform = knownIsoform(related_transcripts)
                    ao.write('\t'.join([gene, ref_super_transcript_id, file, known_isoform]) + '\n')

        eo.close()
        ao.close()
        #jo.close()
        print(a, 'isoform pairs')
        print(len(trans_spliced_isoforms), '>2 trans-spliced isoforms')

        # Iterate back through the original GFF files
        combined_gff = os.path.join(combined_dir, 'combined.gff')
        eo = open(combined_gff, 'w')
        for file in files:
            if os.path.exists(file):
                fn = file
                file = os.path.basename(file)
            else:
                fn = os.path.join(directory, file)
            file = file.split('.g')[0]
            firstRow = True
            exons = []
            isoforms = 2
            with open(fn, 'r') as filepath:
                for line in filepath:
                    data = line.strip()
                    t = data.split('\t')
                    if line[0] =='#':
                        continue
                    chr, a, type, pos1, pos2, b, strand, c, info = t
                    if 'CDS' != type and 'gene' != type:
                        ti = next((i for i, x in enumerate(info.split(';')) if 'transcript_id' in x), -1)
                        if '=' in info:
                            td = '='
                        else:
                            td = '"'
                        transcript_id = info.split(';')[ti].split(td)[1]
                        if (file, transcript_id) in isoforms_to_retain:
                            line = line.replace("PB.", file + '_PB.')
                            eo.write(line)
        eo.close()
        sorted_gff = combined_gff[:-4]+'-sorted.gff'
        try:
            sort_gff(combined_gff, sorted_gff)
        except:
            print ('sort error')
        return combined_dir

def importEnsemblGenes(exon_file):
    global exonCoordinates
    global geneData
    global exonData
    exonCoordinates = {}
    geneData = {}
    gene_chr = {}
    exonData = {}
    strandData = {}
    firstRow = True
    prior_gene = None
    with open(exon_file, 'r') as file:
        for line in file:
            data = line.strip()
            t = data.split('\t')
            if firstRow:
                firstRow = False
            else:
                gene, exon, chr, strand, start, stop = t[:6]
                gene_chr[gene] = chr
                start = int(start)
                stop = int(stop)
                if strand == '-':
                    start, stop = stop, start
                exon_info = (start, stop, exon)
                if gene in geneData:
                    geneData[gene].append(exon_info)
                else:
                    geneData[gene] = [exon_info]
                exonData[gene,exon] = start,stop
                strandData[gene] = strand
                prior_gene=gene
    for gene in geneData:
        geneData[gene].sort()
        if strandData[gene] == '-':
            geneData[gene].reverse()
        index=0
        chr = gene_chr[gene]
        for (start, stop, exon) in geneData[gene]:
            if 'E' in exon:
                exonCoordinates[(chr, start, strandData[gene], 1)] = (gene, exon, index)
                exonCoordinates[(chr, stop, strandData[gene], 2)] = (gene, exon, index)
            index+=1

def sort_gff(concatenated_gff_file, sorted_gff_file):
    def chromosome_sort_key(chromosome):
        chromosome = chromosome.replace('chr', '').replace('Chr', '')
        if chromosome.isdigit():
            return (0, int(chromosome))
        elif chromosome.upper() == 'Y':
            return (1, 0)
        elif chromosome.upper() == 'X':
            return (1, 1)
        else:
            return (2, chromosome)

    # Read the GFF file and store features with their associated transcript start position
    header_lines = []
    features = []
    transcript_starts = {}
    with open(concatenated_gff_file, 'r') as gff:
        for line in gff:
            if line.startswith('#'):
                header_lines.append(line)
            else:
                columns = line.strip().split('\t')
                attributes = {attr.split()[0]: attr.split()[1].strip('"') for attr in columns[8].split(';') if attr}
                transcript_id = attributes.get('transcript_id', None)
                if transcript_id:
                    if transcript_id not in transcript_starts:
                        transcript_starts[transcript_id] = int(columns[3])
                    start_position = transcript_starts[transcript_id]
                    features.append((columns[0], start_position, line))  # (chromosome, transcript start, line)
                else:
                    features.append((columns[0], int(columns[3]), line))  # (chromosome, feature start, line)

    # Sort features by chromosome and transcript start position (or feature start position for non-transcript lines)
    features.sort(key=lambda x: (chromosome_sort_key(x[0]), x[1]))

    # Write the sorted GFF file with headers followed by sorted features
    with open(sorted_gff_file, 'w') as sorted_gff:
        sorted_gff.writelines(header_lines)
        for _, _, line in features:
            sorted_gff.write(line)

def convert_alt_chromosome_ids():
    # Load the mapping file from UCSC Genome Browser
    mapping_url = "https://hgdownload.soe.ucsc.edu/goldenPath/hg38/bigZips/hg38.ncbiRefSeq.gtf.gz"
    mapping_df = pd.read_csv(mapping_url, sep='\t', header=None, comment='#')

    # Extract RefSeq ID to chromosome mapping
    mapping = dict(zip(mapping_df[1], mapping_df[0]))

    def convert_gff(input_gff, output_gff):
        with open(input_gff, 'r') as infile, open(output_gff, 'w') as outfile:
            for line in infile:
                if line.startswith('#'):
                    outfile.write(line)
                    continue
                parts = line.split('\t')
                if parts[0] in mapping:
                    parts[0] = mapping[parts[0]]
                outfile.write('\t'.join(parts))

def reformat_uncoventional_gtf(gtf_file_path,output_file_path):

    # Read the GTF file into a pandas DataFrame
    gtf_df = pd.read_csv(gtf_file_path, sep='\t', header=None, comment='#',
                        names=['seqname', 'source', 'feature', 'start', 'end', 'score', 'strand', 'frame', 'attribute'])

    # Replace 'CDS' with 'exon' in the 'feature' column
    gtf_df['feature'] = gtf_df['feature'].replace('CDS', 'exon')

    # Extract transcript IDs
    gtf_df['transcript_id'] = gtf_df['attribute'].str.extract('transcript_id "([^"]+)"')

    # Identify unique transcripts and their first and last exon positions
    transcript_info = gtf_df.groupby('transcript_id').agg({
        'seqname': 'first',
        'source': 'first',
        'strand': 'first',
        'start': 'min',
        'end': 'max',
        'attribute': 'first'
    }).reset_index()

    # Create transcript entries
    transcript_entries = pd.DataFrame({
        'seqname': transcript_info['seqname'],
        'source': transcript_info['source'],
        'feature': 'transcript',
        'start': transcript_info['start'],
        'end': transcript_info['end'],
        'score': '.',
        'strand': transcript_info['strand'],
        'frame': '.',
        'attribute': transcript_info['attribute'].str.replace(r'(\|[^"]+)', '', regex=True)
    })

    # Add transcript_id to transcript_entries to merge with exons
    transcript_entries['transcript_id'] = transcript_info['transcript_id']

    # Initialize a list to hold final entries
    final_entries = []

    # Iterate over each unique transcript ID to gather transcript and exon entries
    for transcript_id in transcript_info['transcript_id']:
        transcript_entry = transcript_entries[transcript_entries['transcript_id'] == transcript_id]
        exon_entries = gtf_df[gtf_df['transcript_id'] == transcript_id]
        final_entries.append(transcript_entry)
        final_entries.append(exon_entries)

    # Combine all entries into a single DataFrame
    final_df = pd.concat(final_entries).reset_index(drop=True)

    # Drop the temporary 'transcript_id' column
    final_df = final_df.drop(columns=['transcript_id'])

    # Save the reformatted DataFrame to a new GTF file
    final_df.to_csv(output_file_path, sep='\t', header=False, index=False, quoting=3)  # quoting=3 is for quote None

    print(f"Reformatted GTF file with transcript entries saved to {output_file_path}")


if __name__ == '__main__':
    #concatenated = '/Volumes/salomonis2/LabFiles/Frank-Li/neoantigen/revision/blood/Leucegene_AML/MDS-isoforms/head/combined-GFF-alt/combined.gff'
    #sorted_gff = '/Volumes/salomonis2/LabFiles/Frank-Li/neoantigen/revision/blood/Leucegene_AML/MDS-isoforms/head/combined-GFF-alt/sorted_gff_file.gff'
    #sort_gff(concatenated, sorted_gff);sys.exit()

    exon_reference_dir = '/Users/saljh8/Documents/GitHub/altanalyze/AltDatabase/EnsMart91/ensembl/Hs/Hs_Ensembl_exon.txt'
    #exon_reference_dir = '/Volumes/salomonis2/software/AltAnalyze-91/AltAnalyze/AltDatabase/EnsMart91/ensembl/Hs/Hs_Ensembl_exon.txt'
    #exon_reference_dir = '/Users/saljh8/Desktop/Code/AltAnalyze/AltDatabase/EnsMart100/ensembl/Mm/Mm_Ensembl_exon.txt'
    
    gff_input_dir = '/Volumes/salomonis2/LabFiles/Frank-Li/neoantigen/revision/blood/Leucegene_AML/MDS-isoforms/head/'
    gff_input_dir = '/Users/saljh8/Dropbox/Revio/Apharesis-Sequel2/ND167_Iso-Seq/PacBio/Accessory'
    gff_input_dir = '/Users/saljh8/Dropbox/Nanopore/Mouse'
    gff_input_dir = '/Users/saljh8/Downloads/RefSeq'
    #gff_input_dir = '/Volumes/salomonis2/LabFiles/Frank-Li/neoantigen/revision/blood/Leucegene_AML/MDS-isoforms/'
    consolidateLongReadGFFs(gff_input_dir,exon_reference_dir)



import pytest
import pathlib
from altanalyze3.utilities.parser import ArgsParser
from altanalyze3.utilities.helpers import get_md5_sum
from altanalyze3.components.intron_count.main import get_jobs


DATA_FOLDER = pathlib.Path(__file__).resolve().parents[1].joinpath("data")


@pytest.mark.parametrize(
    "args, control_chr",
    [
        (
            [
                "intcount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv",
                "--chr", "chr1", "chr2", "chr3"
            ],
            [
                "chr1", "chr2", "chr3"
            ]
        ),
        (
            [
                "intcount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv",
                "--chr", "1", "2", "3"
            ],
            [
                "chr1", "chr2", "chr3"
            ]
        ),
        (
            [
                "intcount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv",
                "--chr", "chr1", "chr2", "chr3", "chr99"
            ],
            [
                "chr1", "chr2", "chr3"
            ]
        ),
        (
            [
                "intcount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv",
                "--chr", "chr99"
            ],
            []
        ),
        (
            [
                "intcount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv"
            ],
            [
                "chr1", "chr10", "chr11", "chr12", "chr13", "chr14", "chr15",
                "chr16", "chr17", "chr18", "chr19", "chr2", "chr20", "chr21",
                "chr22", "chr3", "chr4", "chr5", "chr6", "chr7", "chr8", "chr9",
                "chrX", "chrY"
            ]
        )
    ]
)
def test_get_jobs(monkeypatch, args, control_chr):
    monkeypatch.chdir(DATA_FOLDER)
    args = ArgsParser(args)
    calculated_chr = [job.contig for job in get_jobs(args)]
    assert sorted(calculated_chr) == sorted(control_chr)


@pytest.mark.parametrize(
    "args, control_md5_sum",
    [
        (
            [
                "intcount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv",
                "--chr", "chr1", "chr2", "chr3"
            ],
            "d314bd03369ef7a019b15b28f695343e"
        ),
        (
            [
                "intcount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv",
                "--chr", "1", "2", "3"
            ],
            "d314bd03369ef7a019b15b28f695343e"
        ),
        (
            [
                "intcount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv",
                "--chr", "chr1", "chr2", "chr3", "chr99"
            ],
            "d314bd03369ef7a019b15b28f695343e"
        ),
        (
            [
                "intcount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv",
                "--chr", "chr99"                   # this should result in an empty file as we forced to use non-existing chromosome
            ],
            "d41d8cd98f00b204e9800998ecf8427e"
        ),
        (
            [
                "intcount",
                "--bam", "Cal27P5-1.bam",
                "--ref", "gene_model_all.tsv"
            ],
            "6f5250f1a841438fadcbe1419966f133"
        ),
        (
            [
                "intcount",
                "--bam", "Cal27P5-1-copy.bam",
                "--ref", "gene_model_all.tsv"
            ],
            "6f5250f1a841438fadcbe1419966f133"
        )
    ]
)
def test_count_introns(monkeypatch, args, control_md5_sum):             # tests all function from intron_count.main file
    monkeypatch.chdir(DATA_FOLDER)
    args = ArgsParser(args)
    args.func(args)
    calculated_md5_sum = get_md5_sum(args.output.with_suffix(".bed"))
    args.output.with_suffix(".bed").unlink()
    assert calculated_md5_sum == control_md5_sum


import time
import random
import string
import logging
import hashlib
import pkg_resources


def get_version():
    """
    Returns current version of the package if it's installed
    """
    pkg = pkg_resources.require("altanalyze3")
    return pkg[0].version if pkg else "unknown version"


def get_tmp_suffix(length=None):
    """
    Returns random filename extension as a string of specified length
    """
    length = 10 if length is None else length
    return "." + "".join(random.choices(string.ascii_uppercase + string.digits, k=length))


def get_md5_sum(location, block_size=2**20):
    md5_sum = hashlib.md5()
    with open(location , "rb") as input_stream:
        while True:
            buf = input_stream.read(block_size)
            if not buf:
                break
            md5_sum.update(buf)
    return md5_sum.hexdigest()

class TimeIt():
    """
    Prints elapsed time to console log (info level)
    """

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, type, value, traceback):
        logging.info (f"""Elapsed time: {round(time.time() - self.start)} sec""")
